import os
import hub
import pathlib
import posixpath
from typing import Dict, Optional, Union, List

from hub.auto.unstructured.kaggle import download_kaggle_dataset
from hub.auto.unstructured.image_classification import ImageClassification
from hub.client.client import HubBackendClient
from hub.client.log import logger
from hub.core.dataset import Dataset, dataset_factory
from hub.core.meta.dataset_meta import DatasetMeta
from hub.util.path import convert_pathlib_to_string_if_needed
from hub.hooks import (
    dataset_created,
    dataset_loaded,
    dataset_written,
    dataset_committed,
)
from hub.constants import (
    DEFAULT_MEMORY_CACHE_SIZE,
    DEFAULT_LOCAL_CACHE_SIZE,
    DEFAULT_READONLY,
    DATASET_META_FILENAME,
)
from hub.util.access_method import (
    check_access_method,
    get_local_dataset,
    parse_access_method,
)
from hub.util.auto import get_most_common_extension
from hub.util.bugout_reporter import feature_report_path, hub_reporter
from hub.util.delete_entry import remove_path_from_backend
from hub.util.keys import dataset_exists
from hub.util.exceptions import (
    AgreementError,
    DatasetHandlerError,
    InvalidFileExtension,
    InvalidPathException,
    PathNotEmptyException,
    SamePathException,
    AuthorizationException,
    UserNotLoggedInException,
    TokenPermissionError,
)
from hub.util.storage import get_storage_and_cache_chain, storage_provider_from_path
from hub.util.compute import get_compute_provider
from hub.util.remove_cache import get_base_storage
from hub.util.cache_chain import generate_chain
from hub.core.storage.hub_memory_object import HubMemoryObject


class dataset:
    @staticmethod
    def init(
        path: Union[str, pathlib.Path],
        read_only: Optional[bool] = None,
        overwrite: bool = False,
        public: bool = False,
        memory_cache_size: int = DEFAULT_MEMORY_CACHE_SIZE,
        local_cache_size: int = DEFAULT_LOCAL_CACHE_SIZE,
        creds: Optional[Union[Dict, str]] = None,
        token: Optional[str] = None,
        verbose: bool = True,
        access_method: str = "stream",
    ):
        """Returns a :class:`~hub.core.dataset.Dataset` object referencing either a new or existing dataset.

        Examples:

            >>> ds = hub.dataset("hub://username/dataset")
            >>> ds = hub.dataset("s3://mybucket/my_dataset")
            >>> ds = hub.dataset("./datasets/my_dataset", overwrite=True)

        Args:
            path (str, pathlib.Path): - The full path to the dataset. Can be:
                - a Hub cloud path of the form ``hub://username/datasetname``. To write to Hub cloud datasets, ensure that you are logged in to Hub (use 'activeloop login' from command line)
                - an s3 path of the form ``s3://bucketname/path/to/dataset``. Credentials are required in either the environment or passed to the creds argument.
                - a local file system path of the form ``./path/to/dataset`` or ``~/path/to/dataset`` or ``path/to/dataset``.
                - a memory path of the form ``mem://path/to/dataset`` which doesn't save the dataset but keeps it in memory instead. Should be used only for testing as it does not persist.
            read_only (bool, optional): Opens dataset in read only mode if this is passed as ``True``. Defaults to ``False``.
                Datasets stored on Hub cloud that your account does not have write access to will automatically open in read mode.
            overwrite (bool): If set to ``True`` this overwrites the dataset if it already exists. Defaults to ``False``.
            public (bool): Defines if the dataset will have public access. Applicable only if Hub cloud storage is used and a new Dataset is being created. Defaults to ``True``.
            memory_cache_size (int): The size of the memory cache to be used in MB.
            local_cache_size (int): The size of the local filesystem cache to be used in MB.
            creds (dict, optional): - A dictionary containing credentials used to access the dataset at the path.
                - If 'aws_access_key_id', 'aws_secret_access_key', 'aws_session_token' are present, these take precedence over credentials present in the environment or in credentials file. Currently only works with s3 paths.
                - It supports 'aws_access_key_id', 'aws_secret_access_key', 'aws_session_token', 'endpoint_url', 'aws_region', 'profile_name' as keys.
            token (str, optional): Activeloop token, used for fetching credentials to the dataset at path if it is a Hub dataset. This is optional, tokens are normally autogenerated.
            verbose (bool): If ``True``, logs will be printed. Defaults to ``True``.
            access_method (str): The access method to use for the dataset. Can be:

                    - 'stream'

                        - Streams the data from the dataset i.e. only fetches data when required. This is the default value.

                    - 'download'

                        - Downloads the data to the local filesystem to the path specified in environment variable ``HUB_DOWNLOAD_PATH``.
                        - Raises an exception if the environment variable is not set, or if the path is not empty.
                        - Will also raise an exception if the dataset does not exist. The 'download' access method can also be modified to specify num_workers and/or scheduler.
                        - For example: 'download:2:processed', will use 2 workers and use processed scheduler, while 'download:3' will use 3 workers and default scheduler (threaded), and 'download:processed' will use a single worker and use processed scheduler.

                    - 'local'

                        - Used when download was already done in a previous run.
                        - Doesn't download the data again.
                        - Raises an exception if ``HUB_DOWNLOAD_PATH`` environment variable is not set or the dataset is not found in ``HUB_DOWNLOAD_PATH``.

        Returns:
            Dataset: Dataset created using the arguments provided.

        Raises:
            AgreementError: When agreement is rejected
            UserNotLoggedInException: When user is not logged in
            InvalidTokenException: If the specified token is invalid
            TokenPermissionError: When there are permission or other errors related to token

        Danger:
            Setting ``overwrite`` to ``True`` will delete all of your data if it exists! Be very careful when setting this parameter.

        Note:
            Any changes made to the dataset in download / local mode will only be made to the local copy and will not be reflected in the original dataset.
        """
        access_method, num_workers, scheduler = parse_access_method(access_method)
        check_access_method(access_method, overwrite)
        path = convert_pathlib_to_string_if_needed(path)

        if creds is None:
            creds = {}

        try:
            storage, cache_chain = get_storage_and_cache_chain(
                path=path,
                read_only=read_only,
                creds=creds,
                token=token,
                memory_cache_size=memory_cache_size,
                local_cache_size=local_cache_size,
            )

            feature_report_path(path, "dataset", {"Overwrite": overwrite}, token=token)
        except Exception as e:
            if isinstance(e, UserNotLoggedInException):
                message = (
                    f"Please log in through the CLI in order to create this dataset, "
                    "or create an API token in the UI and pass it to this method using "
                    "the ‘token’ parameter. The CLI commands are ‘activeloop login’ and "
                    "‘activeloop register."
                )
                raise UserNotLoggedInException(message)
            raise
        ds_exists = dataset_exists(cache_chain)

        if ds_exists:
            if overwrite:
                cache_chain.clear()
                create = True
            else:
                create = False
        else:
            create = True

        try:
            if access_method == "stream":
                ret = dataset_factory(
                    path=path,
                    storage=cache_chain,
                    read_only=read_only,
                    public=public,
                    token=token,
                    verbose=verbose,
                )
                if create:
                    dataset_created(ret)
                else:
                    dataset_loaded(ret)
                return ret

            return get_local_dataset(
                access_method=access_method,
                path=path,
                read_only=read_only,
                memory_cache_size=memory_cache_size,
                local_cache_size=local_cache_size,
                creds=creds,
                token=token,
                verbose=verbose,
                ds_exists=ds_exists,
                num_workers=num_workers,
                scheduler=scheduler,
            )
        except AgreementError as e:
            raise e from None

    @staticmethod
    def exists(
        path: Union[str, pathlib.Path],
        creds: Optional[dict] = None,
        token: Optional[str] = None,
    ) -> bool:
        """Checks if a dataset exists at the given ``path``.

        Args:
            path (str, pathlib.Path): the path which needs to be checked.
            creds (dict, optional): A dictionary containing credentials used to access the dataset at the path.
            token (str, optional): Activeloop token, used for fetching credentials to the dataset at path if it is a Hub dataset. This is optional, tokens are normally autogenerated.

        Returns:
            A boolean confirming whether the dataset exists or not at the given path.
        """
        path = convert_pathlib_to_string_if_needed(path)

        if creds is None:
            creds = {}
        try:
            storage, cache_chain = get_storage_and_cache_chain(
                path=path,
                read_only=True,
                creds=creds,
                token=token,
                memory_cache_size=DEFAULT_MEMORY_CACHE_SIZE,
                local_cache_size=DEFAULT_LOCAL_CACHE_SIZE,
            )
        except (TokenPermissionError):
            # Cloud Dataset does not exist
            return False
        return dataset_exists(storage)

    @staticmethod
    def empty(
        path: Union[str, pathlib.Path],
        overwrite: bool = False,
        public: bool = False,
        memory_cache_size: int = DEFAULT_MEMORY_CACHE_SIZE,
        local_cache_size: int = DEFAULT_LOCAL_CACHE_SIZE,
        creds: Optional[dict] = None,
        token: Optional[str] = None,
        verbose: bool = True,
    ) -> Dataset:
        """Creates an empty dataset

        Args:
            path (str, pathlib.Path): - The full path to the dataset. Can be:
                - a Hub cloud path of the form ``hub://username/datasetname``. To write to Hub cloud datasets, ensure that you are logged in to Hub (use 'activeloop login' from command line)
                - an s3 path of the form ``s3://bucketname/path/to/dataset``. Credentials are required in either the environment or passed to the creds argument.
                - a local file system path of the form ``./path/to/dataset`` or ``~/path/to/dataset`` or ``path/to/dataset``.
                - a memory path of the form ``mem://path/to/dataset`` which doesn't save the dataset but keeps it in memory instead. Should be used only for testing as it does not persist.
            overwrite (bool): If set to ``True`` this overwrites the dataset if it already exists. Defaults to ``False``.
            public (bool): Defines if the dataset will have public access. Applicable only if Hub cloud storage is used and a new Dataset is being created. Defaults to ``False``.
            memory_cache_size (int): The size of the memory cache to be used in MB.
            local_cache_size (int): The size of the local filesystem cache to be used in MB.
            creds (dict, optional): - A dictionary containing credentials used to access the dataset at the path.
                - If 'aws_access_key_id', 'aws_secret_access_key', 'aws_session_token' are present, these take precedence over credentials present in the environment or in credentials file. Currently only works with s3 paths.
                - It supports 'aws_access_key_id', 'aws_secret_access_key', 'aws_session_token', 'endpoint_url', 'aws_region', 'profile_name' as keys.
            token (str, optional): Activeloop token, used for fetching credentials to the dataset at path if it is a Hub dataset. This is optional, tokens are normally autogenerated.
            verbose (bool): If True, logs will be printed. Defaults to True.

        Returns:
            Dataset: Dataset created using the arguments provided.

        Raises:
            DatasetHandlerError: If a Dataset already exists at the given path and overwrite is False.
            UserNotLoggedInException: When user is not logged in
            InvalidTokenException: If the specified toke is invalid
            TokenPermissionError: When there are permission or other errors related to token

        Danger:
            Setting ``overwrite`` to ``True`` will delete all of your data if it exists! Be very careful when setting this parameter.
        """
        path = convert_pathlib_to_string_if_needed(path)

        if creds is None:
            creds = {}

        try:
            storage, cache_chain = get_storage_and_cache_chain(
                path=path,
                read_only=False,
                creds=creds,
                token=token,
                memory_cache_size=memory_cache_size,
                local_cache_size=local_cache_size,
            )

            feature_report_path(path, "empty", {"Overwrite": overwrite}, token=token)
        except Exception as e:
            if isinstance(e, UserNotLoggedInException):
                message = (
                    f"Please log in through the CLI in order to create this dataset, "
                    f"or create an API token in the UI and pass it to this method using the "
                    f"‘token’ parameter. The CLI commands are ‘activeloop login’ and ‘activeloop register’."
                )
                raise UserNotLoggedInException(message)
            raise

        if overwrite and dataset_exists(cache_chain):
            cache_chain.clear()
        elif dataset_exists(cache_chain):
            raise DatasetHandlerError(
                f"A dataset already exists at the given path ({path}). If you want to create"
                f" a new empty dataset, either specify another path or use overwrite=True. "
                f"If you want to load the dataset that exists at this path, use hub.load() instead."
            )
        read_only = storage.read_only
        ret = dataset_factory(
            path=path,
            storage=cache_chain,
            read_only=read_only,
            public=public,
            token=token,
            verbose=verbose,
        )
        dataset_created(ret)
        return ret

    @staticmethod
    def load(
        path: Union[str, pathlib.Path],
        read_only: Optional[bool] = None,
        memory_cache_size: int = DEFAULT_MEMORY_CACHE_SIZE,
        local_cache_size: int = DEFAULT_LOCAL_CACHE_SIZE,
        creds: Optional[dict] = None,
        token: Optional[str] = None,
        verbose: bool = True,
        access_method: str = "stream",
    ) -> Dataset:
        """Loads an existing dataset

        Args:
            path (str, pathlib.Path): - The full path to the dataset. Can be:
                - a Hub cloud path of the form ``hub://username/datasetname``. To write to Hub cloud datasets, ensure that you are logged in to Hub (use 'activeloop login' from command line)
                - an s3 path of the form ``s3://bucketname/path/to/dataset``. Credentials are required in either the environment or passed to the creds argument.
                - a local file system path of the form ``./path/to/dataset`` or ``~/path/to/dataset`` or ``path/to/dataset``.
                - a memory path of the form ``mem://path/to/dataset`` which doesn't save the dataset but keeps it in memory instead. Should be used only for testing as it does not persist.
            read_only (bool, optional): Opens dataset in read only mode if this is passed as ``True``. Defaults to ``False``.
                Datasets stored on Hub cloud that your account does not have write access to will automatically open in read mode.
            memory_cache_size (int): The size of the memory cache to be used in MB.
            local_cache_size (int): The size of the local filesystem cache to be used in MB.
            creds (dict, optional): - A dictionary containing credentials used to access the dataset at the path.
                - If 'aws_access_key_id', 'aws_secret_access_key', 'aws_session_token' are present, these take precedence over credentials present in the environment or in credentials file. Currently only works with s3 paths.
                - It supports 'aws_access_key_id', 'aws_secret_access_key', 'aws_session_token', 'endpoint_url', 'aws_region', 'profile_name' as keys.
            token (str, optional): Activeloop token, used for fetching credentials to the dataset at path if it is a Hub dataset. This is optional, tokens are normally autogenerated.
            verbose (bool): If ``True``, logs will be printed. Defaults to ``True``.
            access_method (str): The access method to use for the dataset. Can be:

                    - 'stream'

                        - Streams the data from the dataset i.e. only fetches data when required. This is the default value.

                    - 'download'

                        - Downloads the data to the local filesystem to the path specified in environment variable `HUB_DOWNLOAD_PATH`.
                        - Raises an exception if the environment variable is not set, or if the path is not empty.
                        - Will also raise an exception if the dataset does not exist. The 'download' access method can also be modified to specify num_workers and/or scheduler.
                        - For example: 'download:2:processed', will use 2 workers and use processed scheduler, while 'download:3' will use 3 workers and default scheduler (threaded), and 'download:processed' will use a single worker and use processed scheduler.

                    - 'local'

                        - Used when download was already done in a previous run.
                        - Doesn't download the data again.
                        - Raises an exception if ``HUB_DOWNLOAD_PATH`` environment variable is not set or the dataset is not found in ``HUB_DOWNLOAD_PATH``.

        Returns:
            Dataset: Dataset loaded using the arguments provided.

        Raises:
            DatasetHandlerError: If a Dataset does not exist at the given path.
            AgreementError: When agreement is rejected
            UserNotLoggedInException: When user is not logged in
            InvalidTokenException: If the specified toke is invalid
            TokenPermissionError: When there are permission or other errors related to token

        Note:
            Any changes made to the dataset in download / local mode will only be made to the local copy and will not be reflected in the original dataset.
        """
        access_method, num_workers, scheduler = parse_access_method(access_method)
        check_access_method(access_method, overwrite=False)
        path = convert_pathlib_to_string_if_needed(path)

        if creds is None:
            creds = {}

        try:
            storage, cache_chain = get_storage_and_cache_chain(
                path=path,
                read_only=read_only,
                creds=creds,
                token=token,
                memory_cache_size=memory_cache_size,
                local_cache_size=local_cache_size,
            )

            feature_report_path(path, "load", {}, token=token)
        except Exception as e:
            if isinstance(e, UserNotLoggedInException):
                message = (
                    "Please log in through the CLI in order to load this dataset, "
                    "or create an API token in the UI and pass it to this method using "
                    "the ‘token’ parameter. The CLI commands are ‘activeloop login’ and "
                    "‘activeloop register’."
                )
                raise UserNotLoggedInException(message)
            raise
        if not dataset_exists(cache_chain):
            raise DatasetHandlerError(
                f"A Hub dataset does not exist at the given path ({path}). Check the path provided or in case you want to create a new dataset, use hub.empty()."
            )

        try:
            if access_method == "stream":
                ret = dataset_factory(
                    path=path,
                    storage=cache_chain,
                    read_only=read_only,
                    token=token,
                    verbose=verbose,
                )
                dataset_loaded(ret)
                return ret
            return get_local_dataset(
                access_method=access_method,
                path=path,
                read_only=read_only,
                memory_cache_size=memory_cache_size,
                local_cache_size=local_cache_size,
                creds=creds,
                token=token,
                verbose=verbose,
                ds_exists=True,
                num_workers=num_workers,
                scheduler=scheduler,
            )
        except AgreementError as e:
            raise e from None

    @staticmethod
    def rename(
        old_path: Union[str, pathlib.Path],
        new_path: Union[str, pathlib.Path],
        creds: Optional[dict] = None,
        token: Optional[str] = None,
    ) -> Dataset:
        """Renames dataset at ``old_path`` to ``new_path``.

        Examples:

            >>> hub.rename("hub://username/image_ds", "hub://username/new_ds")
            >>> hub.rename("s3://mybucket/my_ds", "s3://mybucket/renamed_ds")

        Args:
            old_path (str, pathlib.Path): The path to the dataset to be renamed.
            new_path (str, pathlib.Path): Path to the dataset after renaming.
            creds (dict, optional): - A dictionary containing credentials used to access the dataset at the path.
                - This takes precedence over credentials present in the environment. Currently only works with s3 paths.
                - It supports 'aws_access_key_id', 'aws_secret_access_key', 'aws_session_token', 'endpoint_url' and 'aws_region' as keys.
            token (str, optional): Activeloop token, used for fetching credentials to the dataset at path if it is a Hub dataset. This is optional, tokens are normally autogenerated.

        Returns:
            Dataset: The renamed Dataset.

        Raises:
            DatasetHandlerError: If a Dataset does not exist at the given path or if new path is to a different directory.
        """
        old_path = convert_pathlib_to_string_if_needed(old_path)
        new_path = convert_pathlib_to_string_if_needed(new_path)

        if creds is None:
            creds = {}

        feature_report_path(old_path, "rename", {}, token=token)

        ds = hub.load(old_path, verbose=False, token=token, creds=creds)
        ds.rename(new_path)

        return ds  # type: ignore

    @staticmethod
    def delete(
        path: Union[str, pathlib.Path],
        force: bool = False,
        large_ok: bool = False,
        creds: Optional[dict] = None,
        token: Optional[str] = None,
        verbose: bool = False,
    ) -> None:
        """Deletes a dataset at a given path.

        Args:
            path (str, pathlib.Path): The path to the dataset to be deleted.
            force (bool): Delete data regardless of whether
                it looks like a hub dataset. All data at the path will be removed if set to ``True``.
            large_ok (bool): Delete datasets larger than 1GB. Disabled by default.
            creds (dict, optional): - A dictionary containing credentials used to access the dataset at the path.
                - If 'aws_access_key_id', 'aws_secret_access_key', 'aws_session_token' are present, these take precedence over credentials present in the environment or in credentials file. Currently only works with s3 paths.
                - It supports 'aws_access_key_id', 'aws_secret_access_key', 'aws_session_token', 'endpoint_url', 'aws_region', 'profile_name' as keys.
            token (str, optional): Activeloop token, used for fetching credentials to the dataset at path if it is a Hub dataset. This is optional, tokens are normally autogenerated.
            verbose (bool): If True, logs will be printed. Defaults to True.

        Raises:
            DatasetHandlerError: If a Dataset does not exist at the given path and ``force = False``.
            NotImplementedError: When attempting to delete a managed view.

        Warning:
            This is an irreversible operation. Data once deleted cannot be recovered.
        """
        path = convert_pathlib_to_string_if_needed(path)

        if creds is None:
            creds = {}

        feature_report_path(
            path, "delete", {"Force": force, "Large_OK": large_ok}, token=token
        )

        try:
            qtokens = ["/.queries/", "\\.queries\\"]
            for qt in qtokens:
                if qt in path:
                    raise NotImplementedError(
                        "Deleting managed views by path is not supported. Load the source dataset and do `ds.delete_view(id)` instead."
                    )
            ds = hub.load(path, verbose=False, token=token, creds=creds)
            ds.delete(large_ok=large_ok)
            if verbose:
                logger.info(f"{path} dataset deleted successfully.")
        except Exception as e:
            if force:
                base_storage = storage_provider_from_path(
                    path=path,
                    creds=creds,
                    read_only=False,
                    token=token,
                )
                base_storage.clear()
                remove_path_from_backend(path, token)
                if verbose:
                    logger.info(f"{path} folder deleted successfully.")
            else:
                if isinstance(e, (DatasetHandlerError, PathNotEmptyException)):
                    raise DatasetHandlerError(
                        "A Hub dataset wasn't found at the specified path. "
                        "This may be due to a corrupt dataset or a wrong path. "
                        "If you want to delete the data at the path regardless, use force=True"
                    )
                raise

    @staticmethod
    def like(
        dest: Union[str, pathlib.Path],
        src: Union[str, Dataset, pathlib.Path],
        tensors: Optional[List[str]] = None,
        overwrite: bool = False,
        creds: Optional[dict] = None,
        token: Optional[str] = None,
        public: bool = False,
    ) -> Dataset:
        """Creates a new dataset by copying the ``source`` dataset's structure to a new location. No samples are copied,
        only the meta/info for the dataset and it's tensors.

        Args:
            dest: Empty Dataset or Path where the new dataset will be created.
            src (Union[str, Dataset]): Path or dataset object that will be used as the template for the new dataset.
            tensors (List[str], optional): Names of tensors (and groups) to be replicated. If not specified all tensors in source dataset are considered.
            overwrite (bool): If True and a dataset exists at `destination`, it will be overwritten. Defaults to False.
            creds (dict, optional): - A dictionary containing credentials used to access the dataset at the path.
                - If 'aws_access_key_id', 'aws_secret_access_key', 'aws_session_token' are present, these take precedence over credentials present in the environment or in credentials file. Currently only works with s3 paths.
                - It supports 'aws_access_key_id', 'aws_secret_access_key', 'aws_session_token', 'endpoint_url', 'aws_region', 'profile_name' as keys.
            token (str, optional): Activeloop token, used for fetching credentials to the dataset at path if it is a Hub dataset. This is optional, tokens are normally autogenerated.
            public (bool): Defines if the dataset will have public access. Applicable only if Hub cloud storage is used and a new Dataset is being created. Defaults to False.

        Returns:
            Dataset: New dataset object.
        """
        if isinstance(dest, Dataset):
            path = dest.path
        else:
            path = dest
        feature_report_path(
            path,
            "like",
            {"Overwrite": overwrite, "Public": public, "Tensors": tensors},
            token=token,
        )
        return dataset._like(dest, src, tensors, overwrite, creds, token, public)

    @staticmethod
    def _like(  # (No reporting)
        dest,
        src: Union[str, Dataset],
        tensors: Optional[List[str]] = None,
        overwrite: bool = False,
        creds: Optional[dict] = None,
        token: Optional[str] = None,
        public: bool = False,
        unlink: Union[List[str], bool] = False,
    ) -> Dataset:
        """Copies the `source` dataset's structure to a new location. No samples are copied, only the meta/info for the dataset and it's tensors.

        Args:
            dest: Empty Dataset or Path where the new dataset will be created.
            src (Union[str, Dataset]): Path or dataset object that will be used as the template for the new dataset.
            tensors (List[str], optional): Names of tensors (and groups) to be replicated. If not specified all tensors in source dataset are considered.
            dest (str, pathlib.Path, Dataset): Empty Dataset or Path where the new dataset will be created.
            src (Union[str, pathlib.Path, Dataset]): Path or dataset object that will be used as the template for the new dataset.
            overwrite (bool): If True and a dataset exists at `destination`, it will be overwritten. Defaults to False.
            creds (dict, optional): - A dictionary containing credentials used to access the dataset at the path.
                - If 'aws_access_key_id', 'aws_secret_access_key', 'aws_session_token' are present, these take precedence over credentials present in the environment or in credentials file. Currently only works with s3 paths.
                - It supports 'aws_access_key_id', 'aws_secret_access_key', 'aws_session_token', 'endpoint_url', 'aws_region', 'profile_name' as keys.
            token (str, optional): Activeloop token, used for fetching credentials to the dataset at path if it is a Hub dataset. This is optional, tokens are normally autogenerated.
            public (bool): Defines if the dataset will have public access. Applicable only if Hub cloud storage is used and a new Dataset is being created. Defaults to ``False``.
            unlink (Union[List[str], bool]): List of tensors to be unlinked. If ``True`` passed all tensors will be unlinked. Defaults to ``False``, no tensors are unlinked.

        Returns:
            Dataset: New dataset object.
        """
        dest = convert_pathlib_to_string_if_needed(dest)
        if isinstance(dest, Dataset):
            destination_ds = dest
            dest_path = dest.path
        else:
            dest_path = dest
            destination_ds = dataset.empty(
                dest,
                creds=creds,
                overwrite=overwrite,
                token=token,
                public=public,
            )
        feature_report_path(
            dest_path, "like", {"Overwrite": overwrite, "Public": public}, token=token
        )
        src = convert_pathlib_to_string_if_needed(src)
        if isinstance(src, str):
            source_ds = dataset.load(src)
        else:
            source_ds = src

        if tensors:
            tensors = source_ds._resolve_tensor_list(tensors)  # type: ignore
        else:
            tensors = source_ds.tensors  # type: ignore

        if unlink is True:
            unlink = tensors  # type: ignore
        elif unlink is False:
            unlink = []
        for tensor_name in tensors:  # type: ignore
            if overwrite and tensor_name in destination_ds:
                destination_ds.delete_tensor(tensor_name)
            destination_ds.create_tensor_like(tensor_name, source_ds[tensor_name], unlink=tensor_name in unlink)  # type: ignore

        destination_ds.info.update(source_ds.info.__getstate__())  # type: ignore

        return destination_ds

    @staticmethod
    def copy(
        src: Union[str, pathlib.Path, Dataset],
        dest: Union[str, pathlib.Path],
        tensors: Optional[List[str]] = None,
        overwrite: bool = False,
        src_creds=None,
        src_token=None,
        dest_creds=None,
        dest_token=None,
        num_workers: int = 0,
        scheduler="threaded",
        progressbar=True,
    ):
        """Copies dataset at ``src`` to ``dest``. Version control history is not included.

        Args:
            src (Union[str, Dataset, pathlib.Path]): The Dataset or the path to the dataset to be copied.
            dest (str, pathlib.Path): Destination path to copy to.
            tensors (List[str], optional): Names of tensors (and groups) to be copied. If not specified all tensors are copied.
            overwrite (bool): If True and a dataset exists at ``dest``, it will be overwritten. Defaults to ``False``.
            src_creds (dict, optional): - A dictionary containing credentials used to access the dataset at the path.
                - If 'aws_access_key_id', 'aws_secret_access_key', 'aws_session_token' are present, these take precedence over credentials present in the environment or in credentials file. Currently only works with s3 paths.
                - It supports 'aws_access_key_id', 'aws_secret_access_key', 'aws_session_token', 'endpoint_url', 'aws_region', 'profile_name' as keys.
            src_token (str, optional): Activeloop token, used for fetching credentials to the dataset at ``src`` if it is a Hub dataset. This is optional, tokens are normally autogenerated.
            dest_creds (dict, optional): creds required to create / overwrite datasets at ``dest``.
            dest_token (str, optional): token used to for fetching credentials to ``dest``.
            num_workers (int): The number of workers to use for copying. Defaults to 0. When set to 0, it will always use serial processing, irrespective of the scheduler.
            scheduler (str): The scheduler to be used for copying. Supported values include: 'serial', 'threaded', 'processed' and 'ray'.
                Defaults to 'threaded'.
            progressbar (bool): Displays a progress bar if True (default).

        Returns:
            Dataset: New dataset object.

        Raises:
            DatasetHandlerError: If a dataset already exists at destination path and overwrite is False.
        """

        if isinstance(src, (str, pathlib.Path)):
            src = convert_pathlib_to_string_if_needed(src)
            src_ds = hub.load(src, read_only=True, creds=src_creds, token=src_token)
        else:
            src_ds = src
            src_ds.path = str(src_ds.path)

        dest = convert_pathlib_to_string_if_needed(dest)

        return src_ds.copy(
            dest,
            tensors=tensors,
            overwrite=overwrite,
            creds=dest_creds,
            token=dest_token,
            num_workers=num_workers,
            scheduler=scheduler,
            progressbar=progressbar,
        )

    @staticmethod
    def deepcopy(
        src: Union[str, pathlib.Path],
        dest: Union[str, pathlib.Path],
        tensors: Optional[List[str]] = None,
        overwrite: bool = False,
        src_creds=None,
        src_token=None,
        dest_creds=None,
        dest_token=None,
        num_workers: int = 0,
        scheduler="threaded",
        progressbar=True,
        public: bool = False,
        verbose: bool = True,
    ):
        """Copies dataset at ``src`` to ``dest`` including version control history.

        Args:
            src (str, pathlib.Path): Path to the dataset to be copied.
            dest (str, pathlib.Path): Destination path to copy to.
            tensors (List[str], optional): Names of tensors (and groups) to be copied. If not specified all tensors are copied.
            overwrite (bool): If True and a dataset exists at `destination`, it will be overwritten. Defaults to False.
            src_creds (dict, optional): - A dictionary containing credentials used to access the dataset at the path.
                - If 'aws_access_key_id', 'aws_secret_access_key', 'aws_session_token' are present, these take precedence over credentials present in the environment or in credentials file. Currently only works with s3 paths.
                - It supports 'aws_access_key_id', 'aws_secret_access_key', 'aws_session_token', 'endpoint_url', 'aws_region', 'profile_name' as keys.
            src_token (str, optional): Activeloop token, used for fetching credentials to the dataset at `src` if it is a Hub dataset. This is optional, tokens are normally autogenerated.
            dest_creds (dict, optional): creds required to create / overwrite datasets at ``dest``.
            dest_token (str, optional): token used to for fetching credentials to ``dest``.
            num_workers (int): The number of workers to use for copying. Defaults to 0. When set to 0, it will always use serial processing, irrespective of the scheduler.
            scheduler (str): The scheduler to be used for copying. Supported values include: 'serial', 'threaded', 'processed' and 'ray'.
                Defaults to 'threaded'.
            progressbar (bool): Displays a progress bar if True (default).
            public (bool): Defines if the dataset will have public access. Applicable only if Hub cloud storage is used and a new Dataset is being created. Defaults to ``False``.
            verbose (bool): If True, logs will be printed. Defaults to ``True``.

        Returns:
            Dataset: New dataset object.

        Raises:
            DatasetHandlerError: If a dataset already exists at destination path and overwrite is False.
        """

        src = convert_pathlib_to_string_if_needed(src)
        dest = convert_pathlib_to_string_if_needed(dest)

        report_params = {
            "Overwrite": overwrite,
            "Num_Workers": num_workers,
            "Scheduler": scheduler,
            "Progressbar": progressbar,
            "Public": public,
        }
        if dest.startswith("hub://"):
            report_params["Dest"] = dest
        feature_report_path(src, "deepcopy", report_params, token=dest_token)

        src_ds = hub.load(
            src, read_only=True, creds=src_creds, token=src_token, verbose=False
        )
        src_storage = get_base_storage(src_ds.storage)

        dest_storage, cache_chain = get_storage_and_cache_chain(
            dest,
            creds=dest_creds,
            token=dest_token,
            read_only=False,
            memory_cache_size=DEFAULT_MEMORY_CACHE_SIZE,
            local_cache_size=DEFAULT_LOCAL_CACHE_SIZE,
        )

        if dataset_exists(cache_chain):
            if overwrite:
                cache_chain.clear()
            else:
                raise DatasetHandlerError(
                    f"A dataset already exists at the given path ({dest}). If you want to copy to a new dataset, either specify another path or use overwrite=True."
                )

        metas: Dict[str, DatasetMeta] = {}

        def copy_func(keys, progress_callback=None):
            cache = generate_chain(
                src_storage,
                memory_cache_size=DEFAULT_MEMORY_CACHE_SIZE,
                local_cache_size=DEFAULT_LOCAL_CACHE_SIZE,
                path=src,
            )
            for key in keys:
                val = metas.get(key) or cache[key]
                if isinstance(val, HubMemoryObject):
                    dest_storage[key] = val.tobytes()
                else:
                    dest_storage[key] = val
                if progress_callback:
                    progress_callback(1)

        def copy_func_with_progress_bar(pg_callback, keys):
            copy_func(keys, pg_callback)

        keys = src_storage._all_keys()
        if tensors is not None:
            required_tensors = src_ds._resolve_tensor_list(tensors)
            for t in required_tensors[:]:
                required_tensors.extend(src_ds[t].meta.links)
            required_tensor_paths = set(
                src_ds.meta.tensor_names[t] for t in required_tensors
            )

            all_tensors_in_src = src_ds._tensors()
            all_tensor_paths_in_src = [
                src_ds.meta.tensor_names[t] for t in all_tensors_in_src
            ]
            tensor_paths_to_exclude = [
                t for t in all_tensor_paths_in_src if t not in required_tensor_paths
            ]

            def fltr(k):
                for t in tensor_paths_to_exclude:
                    if k.startswith(t + "/") or "/" + t + "/" in k:
                        return False
                return True

            def keep_group(g):
                for t in tensors:
                    if t == g or t.startswith(g + "/"):
                        return True
                return False

            def process_meta(k):
                if posixpath.basename(k) == DATASET_META_FILENAME:
                    meta = DatasetMeta.frombuffer(src_storage[k])
                    if not meta.tensor_names:  # backward compatibility
                        meta.tensor_names = {t: t for t in meta.tensors}
                    meta.tensors = list(
                        filter(
                            lambda t: meta.tensor_names[t] in required_tensor_paths,
                            meta.tensors,
                        )
                    )
                    meta.hidden_tensors = list(
                        filter(lambda t: t in meta.tensors, meta.hidden_tensors)
                    )
                    meta.groups = list(filter(keep_group, meta.groups))
                    meta.tensor_names = {
                        k: v for k, v in meta.tensor_names.items() if k in meta.tensors
                    }
                    metas[k] = meta
                return k

            keys = filter(fltr, map(process_meta, keys))
        keys = list(keys)
        if tensors:
            assert metas
        len_keys = len(keys)
        if num_workers == 0:
            keys = [keys]
        else:
            keys = [keys[i::num_workers] for i in range(num_workers)]
        compute_provider = get_compute_provider(scheduler, num_workers)
        try:
            if progressbar:
                compute_provider.map_with_progressbar(
                    copy_func_with_progress_bar,
                    keys,
                    len_keys,
                    "Copying dataset",
                )
            else:
                compute_provider.map(copy_func, keys)
        finally:
            compute_provider.close()

        ret = dataset_factory(
            path=dest,
            storage=cache_chain,
            read_only=False,
            public=public,
            token=dest_token,
            verbose=verbose,
        )
        ret._register_dataset()
        dataset_created(ret)
        dataset_written(ret)
        if not ret.has_head_changes:
            dataset_committed(ret)
        return ret

    @staticmethod
    def ingest(
        src: Union[str, pathlib.Path],
        dest: Union[str, pathlib.Path],
        images_compression: str = "auto",
        dest_creds: Optional[Dict] = None,
        progressbar: bool = True,
        summary: bool = True,
        **dataset_kwargs,
    ) -> Dataset:
        """Ingests a dataset from a source and stores it as a structured dataset to destination.

        Args:
            src (str, pathlib.Path): Local path to where the unstructured dataset is stored or path to csv file.
            dest (str, pathlib.Path): - The full path to the dataset. Can be:
                - a Hub cloud path of the form ``hub://username/datasetname``. To write to Hub cloud datasets, ensure that you are logged in to Hub (use 'activeloop login' from command line)
                - an s3 path of the form ``s3://bucketname/path/to/dataset``. Credentials are required in either the environment or passed to the creds argument.
                - a local file system path of the form ``./path/to/dataset`` or ``~/path/to/dataset`` or ``path/to/dataset``.
                - a memory path of the form ``mem://path/to/dataset`` which doesn't save the dataset but keeps it in memory instead. Should be used only for testing as it does not persist.
            images_compression (str): For image classification datasets, this compression will be used for the `images` tensor. If ``images_compression`` is "auto", compression will be automatically determined by the most common extension in the directory.
            dest_creds (Optional[Dict]): A dictionary containing credentials used to access the destination path of the dataset.
            progressbar (bool): Enables or disables ingestion progress bar. Defaults to ``True``.
            summary (bool): If ``True``, a summary of skipped files will be printed after completion. Defaults to ``True``.
            **dataset_kwargs: Any arguments passed here will be forwarded to the dataset creator function.

        Returns:
            Dataset: New dataset object with structured dataset.

        Raises:
            InvalidPathException: If the source directory does not exist.
            SamePathException: If the source and destination path are same.
            AutoCompressionError: If the source director is empty or does not contain a valid extension.
            InvalidFileExtension: If the most frequent file extension is found to be 'None' during auto-compression.

        Note:
            - Currently only local source paths and image classification datasets / csv files are supported for automatic ingestion.
            - Supported filetypes: png/jpeg/jpg/csv.
            - All files and sub-directories with unsupported filetypes are ignored.
            - Valid source directory structures for image classification look like::

                data/
                    img0.jpg
                    img1.jpg
                    ...

            - or::

                data/
                    class0/
                        cat0.jpg
                        ...
                    class1/
                        dog0.jpg
                        ...
                    ...

            - or::

                data/
                    train/
                        class0/
                            img0.jpg
                            ...
                        ...
                    val/
                        class0/
                            img0.jpg
                            ...
                        ...
                    ...

            - Classes defined as sub-directories can be accessed at ``ds["test/labels"].info.class_names``.
            - Support for train and test sub directories is present under ``ds["train/images"]``, ``ds["train/labels"]`` and ``ds["test/images"]``, ``ds["test/labels"]``.
            - Mapping filenames to classes from an external file is currently not supported.
        """
        dest = convert_pathlib_to_string_if_needed(dest)
        feature_report_path(
            dest,
            "ingest",
            {
                "Images_Compression": images_compression,
                "Progressbar": progressbar,
                "Summary": summary,
            },
        )

        src = convert_pathlib_to_string_if_needed(src)

        if isinstance(src, str):
            if os.path.isdir(dest) and os.path.samefile(src, dest):
                raise SamePathException(src)

            if src.endswith(".csv"):
                import pandas as pd  # type:ignore

                if not os.path.isfile(src):
                    raise InvalidPathException(src)
                source = pd.read_csv(src, quotechar='"', skipinitialspace=True)
                ds = dataset.ingest_dataframe(
                    source, dest, dest_creds, progressbar, **dataset_kwargs
                )
                return ds

            if not os.path.isdir(src):
                raise InvalidPathException(src)

            if images_compression == "auto":
                images_compression = get_most_common_extension(src)
                if images_compression is None:
                    raise InvalidFileExtension(src)

            ds = hub.dataset(dest, creds=dest_creds, **dataset_kwargs)

            # TODO: support more than just image classification (and update docstring)
            unstructured = ImageClassification(source=src)

            # TODO: auto detect compression
            unstructured.structure(
                ds,  # type: ignore
                use_progress_bar=progressbar,
                generate_summary=summary,
                image_tensor_args={"sample_compression": images_compression},
            )
        return ds  # type: ignore

    @staticmethod
    def ingest_kaggle(
        tag: str,
        src: Union[str, pathlib.Path],
        dest: Union[str, pathlib.Path],
        exist_ok: bool = False,
        images_compression: str = "auto",
        dest_creds: Optional[Dict] = None,
        kaggle_credentials: dict = None,
        progressbar: bool = True,
        summary: bool = True,
        **dataset_kwargs,
    ) -> Dataset:
        """Download and ingest a kaggle dataset and store it as a structured dataset to destination.

        Args:
            tag (str): Kaggle dataset tag. Example: ``"coloradokb/dandelionimages"`` points to https://www.kaggle.com/coloradokb/dandelionimages
            src (str, pathlib.Path): Local path to where the raw kaggle dataset will be downlaoded to.
            dest (str, pathlib.Path): - The full path to the dataset. Can be:
                - a Hub cloud path of the form ``hub://username/datasetname``. To write to Hub cloud datasets, ensure that you are logged in to Hub (use 'activeloop login' from command line)
                - an s3 path of the form ``s3://bucketname/path/to/dataset``. Credentials are required in either the environment or passed to the creds argument.
                - a local file system path of the form ``./path/to/dataset`` or ``~/path/to/dataset`` or ``path/to/dataset``.
                - a memory path of the form ``mem://path/to/dataset`` which doesn't save the dataset but keeps it in memory instead. Should be used only for testing as it does not persist.
            exist_ok (bool): If the kaggle dataset was already downloaded and ``exist_ok`` is ``True``, ingestion will proceed without error.
            images_compression (str): For image classification datasets, this compression will be used for the ``images`` tensor. If ``images_compression`` is "auto", compression will be automatically determined by the most common extension in the directory.
            dest_creds (Optional[Dict]): A dictionary containing credentials used to access the destination path of the dataset.
            kaggle_credentials (dict): A dictionary containing kaggle credentials {"username":"YOUR_USERNAME", "key": "YOUR_KEY"}. If ``None``, environment variables/the kaggle.json file will be used if available.
            progressbar (bool): Enables or disables ingestion progress bar. Set to ``True`` by default.
            summary (bool): Generates ingestion summary. Set to ``True`` by default.
            **dataset_kwargs: Any arguments passed here will be forwarded to the dataset creator function. See :func:`hub.dataset`.

        Returns:
            Dataset: New dataset object with structured dataset.

        Raises:
            SamePathException: If the source and destination path are same.

        Note:
            Currently only local source paths and image classification datasets are supported for automatic ingestion.
        """
        src = convert_pathlib_to_string_if_needed(src)
        dest = convert_pathlib_to_string_if_needed(dest)

        feature_report_path(
            dest,
            "ingest_kaggle",
            {
                "Images_Compression": images_compression,
                "Exist_Ok": exist_ok,
                "Progressbar": progressbar,
                "Summary": summary,
            },
        )

        if os.path.isdir(src) and os.path.isdir(dest):
            if os.path.samefile(src, dest):
                raise SamePathException(src)

        download_kaggle_dataset(
            tag,
            local_path=src,
            kaggle_credentials=kaggle_credentials,
            exist_ok=exist_ok,
        )

        ds = hub.ingest(
            src=src,
            dest=dest,
            images_compression=images_compression,
            dest_creds=dest_creds,
            progressbar=progressbar,
            summary=summary,
            **dataset_kwargs,
        )

        return ds

    @staticmethod
    def ingest_dataframe(
        src,
        dest: Union[str, pathlib.Path],
        dest_creds: Optional[Dict] = None,
        progressbar: bool = True,
        **dataset_kwargs,
    ):
        """Convert pandas dataframe to a hub Dataset.

        Args:
            src (pd.DataFrame): The pandas dataframe to be converted.
            dest (str, pathlib.Path):
                - The full path to the dataset. Can be:
                - a Hub cloud path of the form ``hub://username/datasetname``. To write to Hub cloud datasets, ensure that you are logged in to Hub (use 'activeloop login' from command line)
                - an s3 path of the form ``s3://bucketname/path/to/dataset``. Credentials are required in either the environment or passed to the creds argument.
                - a local file system path of the form ``./path/to/dataset`` or ``~/path/to/dataset`` or ``path/to/dataset``.
                - a memory path of the form ``mem://path/to/dataset`` which doesn't save the dataset but keeps it in memory instead. Should be used only for testing as it does not persist.
            dest_creds (Optional[Dict]): A dictionary containing credentials used to access the destination path of the dataset.
            progressbar (bool): Enables or disables ingestion progress bar. Set to ``True`` by default.
            **dataset_kwargs: Any arguments passed here will be forwarded to the dataset creator function. See :func:`hub.dataset`.

        Returns:
            Dataset: New dataset created from the dataframe.

        Raises:
            Exception: If ``src`` is not a valid pandas dataframe object.
        """
        import pandas as pd
        from hub.auto.structured.dataframe import DataFrame

        if not isinstance(src, pd.DataFrame):
            raise Exception("Source provided is not a valid pandas dataframe object")

        dest = convert_pathlib_to_string_if_needed(dest)

        ds = hub.dataset(dest, creds=dest_creds, **dataset_kwargs)

        structured = DataFrame(src)
        structured.fill_dataset(ds, progressbar)  # type: ignore
        return ds  # type: ignore

    @staticmethod
    @hub_reporter.record_call
    def list(
        workspace: str = "",
        token: Optional[str] = None,
    ) -> None:
        """List all available hub cloud datasets.

        Args:
            workspace (str): Specify user/organization name. If not given,
                returns a list of all datasets that can be accessed, regardless of what workspace they are in.
                Otherwise, lists all datasets in the given workspace.
            token (str, optional): Activeloop token, used for fetching credentials for Hub datasets. This is optional, tokens are normally autogenerated.

        Returns:
            List: List of dataset names.
        """
        client = HubBackendClient(token=token)
        datasets = client.get_datasets(workspace=workspace)
        return datasets



import pathlib
from hub.core.sample import Sample  # type: ignore
from typing import Optional, Dict, Union

from hub.core.storage.provider import StorageProvider
from hub.util.path import convert_pathlib_to_string_if_needed


def read(
    path: Union[str, pathlib.Path],
    verify: bool = False,
    creds: Optional[Dict] = None,
    compression: Optional[str] = None,
    storage: Optional[StorageProvider] = None,
) -> Sample:
    """Utility that reads raw data from supported files into hub format.

    - Recompresses data into format required by the tensor if permitted by the tensor htype.
    - Simply copies the data in the file if file format matches sample_compression of the tensor, thus maximizing upload speeds.

    Examples:

        >>> ds.create_tensor("images", htype="image", sample_compression="jpeg")
        >>> ds.images.append(hub.read("path/to/cat.jpg"))
        >>> ds.images.shape
        (1, 399, 640, 3)

        >>> ds.create_tensor("videos", htype="video", sample_compression="mp4")
        >>> ds.videos.append(hub.read("path/to/video.mp4"))
        >>> ds.videos.shape
        (1, 136, 720, 1080, 3)

        >>> ds.create_tensor("images", htype="image", sample_compression="jpeg")
        >>> ds.images.append(hub.read("https://picsum.photos/200/300"))
        >>> ds.images[0].shape
        (300, 200, 3)

    Supported file types::

        Image: "bmp", "dib", "gif", "ico", "jpeg", "jpeg2000", "pcx", "png", "ppm", "sgi", "tga", "tiff", "webp", "wmf", "xbm"
        Audio: "flac", "mp3", "wav"
        Video: "mp4", "mkv", "avi"
        Dicom: "dcm"

    Args:
        path (str): Path to a supported file.
        verify (bool):  If True, contents of the file are verified.
        creds (optional, Dict): Credentials for s3, gcp and http urls.
        compression (optional, str): Format of the file. Only required if path does not have an extension.
        storage (optional, StorageProvider): Storage provider to use to retrieve remote files. Useful if multiple files are being read from same storage to minimize overhead of creating a new provider.

    Returns:
        Sample: Sample object. Call ``sample.array`` to get the ``np.ndarray``.

    Note:
        No data is actually loaded until you try to get a property of the returned :class:`Sample`.
        This is useful for passing along to :func:`Tensor.append <hub.core.tensor.Tensor.append>` and :func:`Tensor.extend <hub.core.tensor.Tensor.extend>`.
    """
    path = convert_pathlib_to_string_if_needed(path)
    return Sample(
        path, verify=verify, compression=compression, creds=creds, storage=storage
    )



from io import BytesIO
from PIL import Image  # type: ignore
from hub.core.sample import Sample
import numpy as np


def convert_sample(image_sample: Sample, mode: str) -> Sample:
    if image_sample.path:
        image = Image.open(image_sample.path)
    elif image_sample._buffer:
        image = Image.open(BytesIO(image_sample._buffer))
    if image.mode == mode:
        image.close()
        return image_sample

    image = image.convert(mode)
    image_bytes = BytesIO()
    image.save(image_bytes, format=image_sample.compression)
    converted = Sample(
        buffer=image_bytes.getvalue(), compression=image_sample.compression
    )
    image.close()
    return converted


def to_grayscale(arr: np.ndarray) -> np.ndarray:
    transform = np.array([[[299 / 1000, 587 / 1000, 114 / 1000]]])
    gray = np.sum(transform * arr[:, :, :3], axis=2, dtype=np.uint8)
    return gray


def convert_img_arr(image_arr: np.ndarray, mode: str) -> np.ndarray:
    if len(image_arr.shape) == 2:
        if mode == "L":
            return image_arr
        elif mode == "RGB":
            return np.tile(image_arr[:, :, np.newaxis], (1, 1, 3))

    if (image_arr.shape[-1]) == 4:
        if mode == "L":
            return to_grayscale(image_arr)
        elif mode == "RGB":
            return image_arr[:, :, :3]

    elif (image_arr.shape[-1]) == 3:
        if mode == "L":
            return to_grayscale(image_arr)
        elif mode == "RGB":
            return image_arr

    raise ValueError("Invalid image")



from datetime import datetime
from hub.client.utils import get_user_name
from typing import List, Optional


class CommitNode:
    """Contains all the Version Control information about a particular commit."""

    def __init__(self, branch: str, commit_id: str):
        self.commit_id = commit_id
        self.branch = branch
        self.children: List["CommitNode"] = []
        self.parent: Optional["CommitNode"] = None
        self.commit_message: Optional[str] = None
        self.commit_time: Optional[datetime] = None
        self.commit_user_name: Optional[str] = None
        self.merge_parent: Optional["CommitNode"] = None

    def add_child(self, node: "CommitNode"):
        """Adds a child to the node, used for branching."""
        node.parent = self
        self.children.append(node)

    def copy(self):
        node = CommitNode(self.branch, self.commit_id)
        node.commit_message = self.commit_message
        node.commit_user_name = self.commit_user_name
        node.commit_time = self.commit_time
        return node

    def add_successor(self, node: "CommitNode", message: Optional[str] = None):
        """Adds a successor (a type of child) to the node, used for commits."""
        node.parent = self
        self.children.append(node)
        self.commit_message = message
        self.commit_user_name = get_user_name()
        self.commit_time = datetime.utcnow()

    def merge_from(self, node: "CommitNode"):
        """Merges the given node into this node."""
        self.merge_parent = node

    @property
    def is_merge_node(self):
        """Returns True if the node is a merge node."""
        return self.merge_parent is not None

    def __repr__(self) -> str:
        return f"Commit : {self.commit_id} ({self.branch}) \nAuthor : {self.commit_user_name}\nTime   : {str(self.commit_time)[:-7]}\nMessage: {self.commit_message}"

    @property
    def is_head_node(self) -> bool:
        """Returns True if the node is the head node of the branch."""
        return self.commit_time is None

    __str__ = __repr__



"""
Tensorflow dataset wrapper
"""
import abc

import tensorflow as tf  # type: ignore


def hub_tf_adapter(_fn):
    """
    Decorator function
    """

    def inner(*args, **kwargs):
        return DatasetAdapter(_fn(*args, **kwargs))

    return inner


# pylint: disable=too-many-public-methods
class HubTensorflowDataset(tf.data.Dataset):
    """Represents a potentially large set of elements.

    A `HubTensorflowDataset` can be used to represent an input pipeline as a
    collection of elements and a "logical plan" of transformations that act on
    those elements.
    """

    def __init__(self):
        try:
            variant_tensor = self._as_variant_tensor()
        except AttributeError as attr_ex:
            if "_as_variant_tensor" in str(attr_ex):
                raise AttributeError(
                    "Please use _variant_tensor instead of "
                    "_as_variant_tensor() to obtain the variant "
                    "associated with a dataset"
                ) from attr_ex
        super(__class__, self).__init__(variant_tensor)

    @abc.abstractmethod
    def _as_variant_tensor(self):
        """Creates a scalar `tf.Tensor` of `tf.variant` representing this dataset.
        Returns:
          A scalar `tf.Tensor` of `tf.variant` type, which represents this dataset.
        """
        raise NotImplementedError("Dataset._as_variant_tensor")

    @staticmethod
    @hub_tf_adapter
    def from_tensors(tensors):
        return tf.data.Dataset.from_tensors(tensors)

    @staticmethod
    @hub_tf_adapter
    def from_tensor_slices(tensors):
        return tf.data.Dataset.from_tensor_slices(tensors)

    @staticmethod
    @hub_tf_adapter
    def from_generator(*args, **kwargs):
        return tf.data.Dataset.from_generator(*args, **kwargs)

    @staticmethod
    @hub_tf_adapter
    def range(*args, **kwargs):
        return tf.data.Dataset.range(*args, **kwargs)

    @staticmethod
    @hub_tf_adapter
    def zip(datasets):
        return tf.data.Dataset.zip(datasets)

    @hub_tf_adapter
    def concatenate(self, *args, **kwargs):
        return super(__class__, self).concatenate(*args, **kwargs)

    @hub_tf_adapter
    def prefetch(self, *args, **kwargs):
        return super(__class__, self).prefetch(*args, **kwargs)

    @staticmethod
    @hub_tf_adapter
    def list_files(*args, **kwargs):
        return tf.data.Dataset.list_files(*args, **kwargs)

    @hub_tf_adapter
    def repeat(self, *args, **kwargs):
        return super(__class__, self).repeat(*args, **kwargs)

    @hub_tf_adapter
    def shuffle(self, *args, **kwargs):
        return super(__class__, self).shuffle(*args, **kwargs)

    @hub_tf_adapter
    def cache(self, *args, **kwargs):
        return super(__class__, self).cache(*args, **kwargs)

    @hub_tf_adapter
    def take(self, *args, **kwargs):
        return super(__class__, self).take(*args, **kwargs)

    @hub_tf_adapter
    def skip(self, *args, **kwargs):
        return super(__class__, self).skip(*args, **kwargs)

    @hub_tf_adapter
    def shard(self, *args, **kwargs):
        return super(__class__, self).shard(*args, **kwargs)

    @hub_tf_adapter
    def batch(self, *args, **kwargs):
        return super(__class__, self).batch(*args, **kwargs)

    @hub_tf_adapter
    def padded_batch(self, *args, **kwargs):
        return super(__class__, self).padded_batch(*args, **kwargs)

    @hub_tf_adapter
    def map(self, *args, **kwargs):
        return super(__class__, self).map(*args, **kwargs)

    @hub_tf_adapter
    def flat_map(self, *args, **kwargs):
        return super(__class__, self).flat_map(*args, **kwargs)

    @hub_tf_adapter
    def interleave(self, *args, **kwargs):
        return super(__class__, self).interleave(*args, **kwargs)

    @hub_tf_adapter
    def filter(self, *args, **kwargs):
        return super(__class__, self).filter(*args, **kwargs)

    @hub_tf_adapter
    def apply(self, *args, **kwargs):
        return super(__class__, self).apply(*args, **kwargs)

    @hub_tf_adapter
    def window(self, *args, **kwargs):
        return super(__class__, self).window(*args, **kwargs)

    @hub_tf_adapter
    def unbatch(self):
        return super(__class__, self).unbatch()

    @hub_tf_adapter
    def with_options(self, *args, **kwargs):
        return super(__class__, self).with_options(*args, **kwargs)


class DatasetAdapter(HubTensorflowDataset):
    """Wraps a V2 `Dataset` object in the `HubTensorflowDataset` API."""

    # pylint: disable=abstract-method
    def __init__(self, dataset):
        self._dataset = dataset
        super(__class__, self).__init__()

    def _as_variant_tensor(self):
        return self._dataset._variant_tensor  # pylint: disable=protected-access

    def _inputs(self):
        return self._dataset._inputs()  # pylint: disable=protected-access

    def _functions(self):
        return self._dataset._functions()  # pylint: disable=protected-access

    def options(self):
        return self._dataset.options()

    @property
    def element_spec(self):
        return self._dataset.element_spec  # pylint: disable=protected-access

    def __iter__(self):
        return iter(self._dataset)



import os
import struct
from typing import List, Optional, Union
from hub.core.compression import decompress_array, decompress_bytes
from hub.core.sample import Sample  # type: ignore
from hub.core.serialize import (
    check_sample_shape,
    bytes_to_text,
)
from hub.core.tiling.sample_tiles import SampleTiles
from hub.util.exceptions import EmptyTensorError
from hub.util.video import normalize_index
from .base_chunk import BaseChunk, InputSample
import numpy as np


class SampleCompressedChunk(BaseChunk):
    def extend_if_has_space(self, incoming_samples: List[InputSample], update_tensor_meta: bool = True) -> float:  # type: ignore
        self.prepare_for_write()
        num_samples: float = 0
        dtype = self.dtype if self.is_byte_compression else None
        compr = self.compression

        for i, incoming_sample in enumerate(incoming_samples):
            serialized_sample, shape = self.serialize_sample(incoming_sample, compr)
            if shape is not None:
                self.num_dims = self.num_dims or len(shape)
                check_sample_shape(shape, self.num_dims)

            if isinstance(serialized_sample, SampleTiles):
                incoming_samples[i] = serialized_sample  # type: ignore
                if self.is_empty:
                    self.write_tile(serialized_sample)
                    num_samples += 0.5
                break

            else:
                sample_nbytes = len(serialized_sample)
                if self.is_empty or self.can_fit_sample(sample_nbytes):
                    self.data_bytes += serialized_sample  # type: ignore

                    self.register_in_meta_and_headers(
                        sample_nbytes,
                        shape,
                        update_tensor_meta=update_tensor_meta,
                    )
                    num_samples += 1
                else:
                    if serialized_sample:
                        buffer = serialized_sample
                        sample = Sample(
                            buffer=buffer, compression=compr, shape=shape, dtype=dtype  # type: ignore
                        )
                        incoming_samples[i] = sample
                    break
        return num_samples

    def read_sample(  # type: ignore
        self,
        local_index: int,
        cast: bool = True,
        copy: bool = False,
        sub_index: Optional[Union[int, slice]] = None,
        stream: bool = False,
        decompress: bool = True,
        is_tile: bool = False,
    ):
        if self.is_empty_tensor:
            raise EmptyTensorError(
                "This tensor has only been populated with empty samples. "
                "Need to add at least one non-empty sample before retrieving data."
            )

        partial_sample_tile = self._get_partial_sample_tile()
        if partial_sample_tile is not None:
            return partial_sample_tile
        buffer = self.memoryview_data
        if not self.byte_positions_encoder.is_empty():
            sb, eb = self.byte_positions_encoder[local_index]
            if stream and self.is_video_compression:
                header_size = struct.unpack("<i", buffer[-4:])[
                    0
                ]  # last 4 bytes store size of header

                # create subfile url to pass to ffmpeg. header_size + sb will give starting point of video bytes.
                # https://ffmpeg.org/ffmpeg-protocols.html#subfile
                buffer = (
                    f"subfile,,start,{header_size + sb},end,{header_size + eb},,:"
                    + bytes(buffer[:-4]).decode("utf-8")
                )
                if not decompress:
                    return buffer
            else:
                buffer = buffer[sb:eb]
        if not decompress:
            return bytes(buffer) if copy else buffer
        if not is_tile and self.is_fixed_shape:
            shape = tuple(self.tensor_meta.min_shape)
        else:
            shape = self.shapes_encoder[local_index]
        nframes = shape[0]
        if self.is_text_like:
            buffer = decompress_bytes(buffer, compression=self.compression)
            buffer = bytes(buffer)
            return bytes_to_text(buffer, self.htype)

        squeeze = isinstance(sub_index, int)

        start, stop, step, reverse = normalize_index(sub_index, nframes)

        if start > nframes:
            raise IndexError("Start index out of bounds.")

        sample = decompress_array(
            buffer,
            shape,
            self.dtype,
            self.compression,
            start_idx=start,
            end_idx=stop,
            step=step,
            reverse=reverse,
        )

        if squeeze:
            sample = sample.squeeze(0)

        if cast and sample.dtype != self.dtype:
            sample = sample.astype(self.dtype)
        elif copy and not sample.flags["WRITEABLE"]:
            sample = sample.copy()
        return sample

    def update_sample(self, local_index: int, sample: InputSample):
        self.prepare_for_write()
        serialized_sample, shape = self.serialize_sample(
            sample, self.compression, break_into_tiles=False
        )

        self.check_shape_for_update(shape)
        old_data = self.data_bytes
        self.data_bytes = self.create_updated_data(
            local_index, old_data, serialized_sample
        )

        # update encoders and meta
        new_nb = (
            None if self.byte_positions_encoder.is_empty() else len(serialized_sample)
        )
        self.update_in_meta_and_headers(local_index, new_nb, shape)



from typing import Callable, Dict, List, Optional, Union
from hub.experimental.convert_to_hub3 import dataset_to_hub3, verify_base_storage  # type: ignore
from hub.experimental.util import raise_indra_installation_error  # type: ignore
from hub.experimental.util import collate_fn as default_collate  # type: ignore
from hub.experimental.hub3_query import query
from hub.integrations.pytorch.common import PytorchTransformFunction
from hub.util.bugout_reporter import hub_reporter
from hub.util.dataset import map_tensor_keys

try:
    from indra import Loader  # type: ignore

    INDRA_INSTALLED = True
except ImportError:
    INDRA_INSTALLED = False


class Hub3DataLoader:
    def __init__(
        self,
        dataset,
        _batch_size=None,
        _shuffle=None,
        _num_threads=None,
        _num_workers=None,
        _collate=None,
        _transform=None,
        _distributed=None,
        _prefetch_factor=None,
        _tensors=None,
        _drop_last=False,
        _mode=None,
        _return_index=None,
    ):
        raise_indra_installation_error(INDRA_INSTALLED)
        self.dataset = dataset
        self._batch_size = _batch_size
        self._shuffle = _shuffle
        self._num_threads = _num_threads
        self._num_workers = _num_workers
        self._collate = _collate
        self._transform = _transform
        self._distributed = _distributed
        self._prefetch_factor = _prefetch_factor
        self._tensors = _tensors
        self._drop_last = _drop_last
        self._mode = _mode
        self._return_index = _return_index

    def batch(self, batch_size: int, drop_last: bool = False):
        """Returns a batched hub.experimental.Hub3DataLoader object.


        Args:
            batch_size (int): Number of samples in each batch.
            drop_last (bool): If True, the last batch will be dropped if its size is less than batch_size. Defaults to False.


        Returns:
            Hub3DataLoader: A hub.experimental.Hub3DataLoader object.


        Raises:
            ValueError: If .batch() has already been called.
        """
        if self._batch_size is not None:
            raise ValueError("batch size is already set")

        all_vars = self.__dict__.copy()
        all_vars["_batch_size"] = batch_size
        all_vars["_drop_last"] = drop_last
        return self.__class__(**all_vars)

    def shuffle(self):
        """Returns a shuffled hub.experimental.Hub3DataLoader object.


        Returns:
            Hub3DataLoader: A hub.experimental.Hub3DataLoader object.


        Raises:
            ValueError: If .shuffle() has already been called.
        """
        if self._shuffle is not None:
            raise ValueError("shuffle is already set")
        all_vars = self.__dict__.copy()
        all_vars["_shuffle"] = True
        return self.__class__(**all_vars)

    def transform(self, transform: Union[Callable, Dict[str, Optional[Callable]]]):
        """Returns a transformed hub.experimental.Hub3DataLoader object.


        Args:
            transform (Callable or Dict[Callable]): A function or dictionary of functions to apply to the data.


        Returns:
            Hub3DataLoader: A hub.experimental.Hub3DataLoader object.


        Raises:
            ValueError: If .transform() has already been called.
        """
        if self._transform is not None:
            raise ValueError("transform is already set")
        all_vars = self.__dict__.copy()
        if isinstance(transform, dict):
            tensors = [k for k in transform.keys() if k != "index"]
            tensors = map_tensor_keys(self.dataset, tensors)
            if self._tensors:
                raise ValueError(
                    f"Tensors have already been specified in the .{self._mode} method."
                )
            all_vars["_tensors"] = map_tensor_keys(self.dataset, tensors)
            transform = PytorchTransformFunction(transform_dict=transform)
        else:
            transform = PytorchTransformFunction(composite_transform=transform)
        all_vars["_transform"] = transform
        return self.__class__(**all_vars)

    def query(self, query_string: str):
        """Returns a sliced hub.experimental.Hub3DataLoader object with given query results.
        It allows to run SQL like queries on dataset and extract results. Currently supported keywords are the following:

        +-------------------------------------------+
        | SELECT                                    |
        +-------------------------------------------+
        | FROM                                      |
        +-------------------------------------------+
        | CONTAINS                                  |
        +-------------------------------------------+
        | ORDER BY                                  |
        +-------------------------------------------+
        | GROUP BY                                  |
        +-------------------------------------------+
        | LIMIT                                     |
        +-------------------------------------------+
        | OFFSET                                    |
        +-------------------------------------------+
        | RANDOM() -> for shuffling query results   |
        +-------------------------------------------+


        Args:
            query_string (str): An SQL string adjusted with new functionalities to run on the dataset object

        Returns:
            Hub3DataLoader: A hub.experimental.Hub3DataLoader object.

        Examples:
            >>> import hub
            >>> from hub.experimental import dataloader
            >>> ds = hub.load('hub://activeloop/fashion-mnist-train')
            >>> query_ds_train = dataloader(ds_train).query("select * where labels != 5")

            >>> import hub
            >>> from hub.experimental import query
            >>> ds_train = hub.load('hub://activeloop/coco-train')
            >>> query_ds_train = dataloader(ds_train).query("(select * where contains(categories, 'car') limit 1000) union (select * where contains(categories, 'motorcycle') limit 1000)")
        """
        all_vars = self.__dict__.copy()
        all_vars["dataset"] = query(self.dataset, query_string)
        return self.__class__(**all_vars)

    @hub_reporter.record_call
    def pytorch(
        self,
        num_workers: int = 0,
        collate_fn: Optional[Callable] = None,
        tensors: Optional[List[str]] = None,
        num_threads: Optional[int] = None,
        prefetch_factor: int = 10,
        distributed: bool = False,
        return_index: bool = True,
    ):
        """Returns a hub.experimental.Hub3DataLoader object.


        Args:
            num_workers (int): Number of workers to use for transforming and processing the data. Defaults to 0.
            collate_fn (Callable, Optional): merges a list of samples to form a mini-batch of Tensor(s).
            tensors (List[str], Optional): List of tensors to load. If None, all tensors are loaded. Defaults to None.
            num_threads (int, Optional): Number of threads to use for fetching and decompressing the data. If None, the number of threads is automatically determined. Defaults to None.
            prefetch_factor (int): Number of batches to transform and collate in advance per worker. Defaults to 10.
            distributed (bool): Used for DDP training. Distributes different sections of the dataset to different ranks. Defaults to False.
            return_index (bool): Used to idnetify where loader needs to retur sample index or not. Defaults to True.


        Returns:
            Hub3DataLoader: A hub.experimental.Hub3DataLoader object.


        Raises:
            ValueError: If .to_pytorch() or .to_numpy() has already been called.
        """
        if self._mode is not None:
            if self._mode == "numpy":
                raise ValueError("Can't call .to_pytorch after .to_numpy()")
            raise ValueError("already called .to_pytorch()")
        all_vars = self.__dict__.copy()
        all_vars["_num_workers"] = num_workers
        all_vars["_collate"] = collate_fn
        if tensors:
            if "index" in tensors:
                raise ValueError(
                    "index is not a tensor, to get index, pass return_index=True"
                )
            tensors = map_tensor_keys(self.dataset, tensors)
            if self._tensors:
                raise ValueError(
                    "Tensors have already been specified by passing a dictionary to .transform() method"
                )
        all_vars["_tensors"] = self._tensors or tensors
        all_vars["_num_threads"] = num_threads
        all_vars["_prefetch_factor"] = prefetch_factor
        all_vars["_distributed"] = distributed
        all_vars["_return_index"] = return_index
        all_vars["_mode"] = "pytorch"
        return self.__class__(**all_vars)

    @hub_reporter.record_call
    def numpy(
        self,
        num_workers: int = 0,
        tensors: Optional[List[str]] = None,
        num_threads: Optional[int] = None,
        prefetch_factor: int = 10,
    ):
        """Returns a hub.experimental.Hub3DataLoader object.

        Args:
            num_workers (int): Number of workers to use for transforming and processing the data. Defaults to 0.
            tensors (List[str], Optional): List of tensors to load. If None, all tensors are loaded. Defaults to None.
            num_threads (int, Optional): Number of threads to use for fetching and decompressing the data. If None, the number of threads is automatically determined. Defaults to None.
            prefetch_factor (int): Number of batches to transform and collate in advance per worker. Defaults to 10.


        Returns:
            Hub3DataLoader: A hub.experimental.Hub3DataLoader object.


        Raises:
            ValueError: If .to_pytorch() or .to_numpy() has already been called.
        """
        if self._mode is not None:
            if self._mode == "pytorch":
                raise ValueError("Can't call .to_numpy after .to_pytorch()")
            raise ValueError("already called .to_numpy()")
        all_vars = self.__dict__.copy()
        all_vars["_num_workers"] = num_workers
        if tensors:
            if "index" in tensors:
                raise ValueError(
                    "index is not a tensor, to get index, pass return_index=True"
                )
            tensors = map_tensor_keys(self.dataset, tensors)
            if self._tensors:
                raise ValueError(
                    "Tensors have already been specified by passing a dictionary to .transform() method"
                )
        all_vars["_tensors"] = self._tensors or tensors
        all_vars["_num_threads"] = num_threads
        all_vars["_prefetch_factor"] = prefetch_factor
        all_vars["_mode"] = "numpy"
        return self.__class__(**all_vars)

    def __iter__(self):
        dataset = dataset_to_hub3(self.dataset)
        batch_size = self._batch_size or 1
        drop_last = self._drop_last or False
        return_index = self._return_index
        if return_index is None:
            return_index = True

        shuffle = self._shuffle or False

        transform_fn = self._transform

        num_workers = self._num_workers or 0
        if self._collate is None and self._mode == "pytorch":
            collate_fn = default_collate
        else:
            collate_fn = self._collate
        tensors = self._tensors or map_tensor_keys(self.dataset, None)
        num_threads = self._num_threads
        prefetch_factor = self._prefetch_factor
        distributed = self._distributed or False
        upcast = (
            self._mode == "pytorch"
        )  # only upcast for pytorch, this handles unsupported dtypes
        return iter(
            Loader(
                dataset,
                batch_size=batch_size,
                num_threads=num_threads,
                shuffle=shuffle,
                num_workers=num_workers,
                collate_fn=collate_fn,
                transform_fn=transform_fn,
                distributed=distributed,
                prefetch_factor=prefetch_factor,
                tensors=tensors,
                drop_last=drop_last,
                upcast=upcast,
                return_index=return_index,
            )
        )


def dataloader(dataset) -> Hub3DataLoader:
    """Returns a hub.experimental.Hub3DataLoader object which can be transformed to either pytorch dataloader or numpy.


    Args:
        dataset: hub.Dataset object on which dataloader needs to be built

    Returns:
        Hub3DataLoader: A hub.experimental.Hub3DataLoader object.


    Examples:
        >>> import hub
        >>> from hub.experimental import dataloader
        >>> ds_train = hub.load('hub://activeloop/fashion-mnist-train')
        >>> train_loader = dataloader(ds_train).numpy()
        >>> for i, data in enumerate(train_loader):
        ...     # custom logic on data
        ...     pass

        >>> import torch
        >>> from torchvision import datasets, transforms, models
        >>> ds_train = hub.load('hub://activeloop/fashion-mnist-train')
        >>> tform = transforms.Compose([
        ...     transforms.ToPILImage(), # Must convert to PIL image for subsequent operations to run
        ...     transforms.RandomRotation(20), # Image augmentation
        ...     transforms.ToTensor(), # Must convert to pytorch tensor for subsequent operations to run
        ...     transforms.Normalize([0.5], [0.5]),
        ... ])
        >>> batch_size = 32
        >>> train_loader = dataloader(ds_train)
        ...     .transform({'images': tform, 'labels': None})
        ...     .batch(batch_size).pytorch()
        >>> for i, data in enumerate(train_loader):
        ...     # custom logic on data
        ...     pass
    """
    verify_base_storage(dataset)
    return Hub3DataLoader(dataset)



from hub.core.storage.gcs import GCSProvider
from hub.core.storage.google_drive import GDriveProvider
from hub.util.storage import storage_provider_from_hub_path
from hub.core.storage.s3 import S3Provider
from hub.core.storage.local import LocalProvider
import os
import hub
from hub.constants import (
    HUB_CLOUD_OPT,
    KEEP_STORAGE_OPT,
    LOCAL_OPT,
    MEMORY_OPT,
    PYTEST_GCS_PROVIDER_BASE_ROOT,
    PYTEST_S3_PROVIDER_BASE_ROOT,
    PYTEST_HUB_CLOUD_PROVIDER_BASE_ROOT,
    PYTEST_LOCAL_PROVIDER_BASE_ROOT,
    PYTEST_MEMORY_PROVIDER_BASE_ROOT,
    S3_OPT,
    GCS_OPT,
    GDRIVE_OPT,
    S3_PATH_OPT,
    GDRIVE_PATH_OPT,
    ENV_GOOGLE_APPLICATION_CREDENTIALS,
    ENV_GDRIVE_CLIENT_ID,
    ENV_GDRIVE_CLIENT_SECRET,
    ENV_GDRIVE_REFRESH_TOKEN,
)
import posixpath
from hub.tests.common import (
    SESSION_ID,
    current_test_name,
    get_dummy_data_path,
    is_opt_true,
)
import pytest
import sys


MEMORY = "memory"
LOCAL = "local"
S3 = "s3"
GDRIVE = "gdrive"
GCS = "gcs"
HUB_CLOUD = "hub_cloud"

_GIT_CLONE_CACHE_DIR = ".test_resources"

_HUB_TEST_RESOURCES_URL = "https://www.github.com/activeloopai/hub-test-resources.git"
_PILLOW_URL = "https://www.github.com/python-pillow/Pillow.git"


def _repo_name_from_git_url(url):
    repo_name = posixpath.split(url)[-1]
    repo_name = repo_name.split("@", 1)[0]
    if repo_name.endswith(".git"):
        repo_name = repo_name[:-4]
    return repo_name


def _git_clone(url):
    _repo_name = _repo_name_from_git_url(url)
    cached_dir = _GIT_CLONE_CACHE_DIR + "/" + _repo_name
    if not os.path.isdir(cached_dir):
        if not os.path.isdir(_GIT_CLONE_CACHE_DIR):
            os.mkdir(_GIT_CLONE_CACHE_DIR)
        cwd = os.getcwd()
        os.chdir(_GIT_CLONE_CACHE_DIR)
        try:
            os.system(f"git clone " + url)
        finally:
            os.chdir(cwd)
    assert os.path.isdir(cached_dir)
    return cached_dir


def _download_hub_test_images():
    path = _git_clone(_HUB_TEST_RESOURCES_URL)
    jpeg_path = path + "/images/jpeg"
    return [os.path.join(jpeg_path, f) for f in os.listdir(jpeg_path)]


def _download_hub_test_videos():
    path = _git_clone(_HUB_TEST_RESOURCES_URL)
    mp4_path = path + "/videos/mp4"
    return [os.path.join(mp4_path, f) for f in os.listdir(mp4_path)]


def _download_pil_test_images(ext=[".jpg", ".png"]):
    paths = {e: [] for e in ext}
    corrupt_file_keys = [
        "broken",
        "_dos",
        "truncated",
        "chunk_no_fctl",
        "syntax_num_frames_zero",
    ]

    path = _git_clone(_PILLOW_URL)
    dirs = [
        path + x
        for x in [
            "/Tests/images",
            "/Tests/images/apng",
            "/Tests/images/imagedraw",
        ]
    ]
    for d in dirs:
        for f in os.listdir(d):
            brk = False
            for k in corrupt_file_keys:
                if k in f:
                    brk = True
                    break
            if brk:
                continue
            for e in ext:
                if f.lower().endswith(e):
                    paths[e].append(os.path.join(d, f))
                    break
    return paths


def _get_path_composition_configs(request):
    return {
        MEMORY: {
            "base_root": PYTEST_MEMORY_PROVIDER_BASE_ROOT,
            "use_id": False,
            "is_id_prefix": False,
            # if is_id_prefix (and use_id=True), the session id comes before test name, otherwise it is reversed
            "use_underscores": False,
        },
        LOCAL: {
            "base_root": PYTEST_LOCAL_PROVIDER_BASE_ROOT,
            "use_id": False,
            "is_id_prefix": False,
            "use_underscores": False,
        },
        S3: {
            "base_root": request.config.getoption(S3_PATH_OPT),
            "use_id": True,
            "is_id_prefix": True,
            "use_underscores": False,
        },
        GDRIVE: {
            "base_root": request.config.getoption(GDRIVE_PATH_OPT),
            "use_id": True,
            "is_id_prefix": True,
            "use_underscores": False,
        },
        GCS: {
            "base_root": PYTEST_GCS_PROVIDER_BASE_ROOT,
            "use_id": True,
            "is_id_prefix": True,
            "use_underscores": False,
        },
        HUB_CLOUD: {
            "base_root": PYTEST_HUB_CLOUD_PROVIDER_BASE_ROOT,
            "use_id": True,
            "is_id_prefix": True,
            "use_underscores": True,
        },
    }


def _get_storage_path(
    request, storage_name, with_current_test_name=True, info_override={}
):
    info = _get_path_composition_configs(request)[storage_name]
    info.update(info_override)

    root = info["base_root"]

    path = ""
    if with_current_test_name:
        path = current_test_name()

    if info["use_id"]:
        if info["is_id_prefix"]:
            path = posixpath.join(SESSION_ID, path)
        else:
            path = posixpath.join(path, SESSION_ID)

    if info["use_underscores"]:
        path = path.replace("/", "_")

    root = posixpath.join(root, path).strip("/")
    return root


@pytest.fixture
def memory_path(request):
    if is_opt_true(request, MEMORY_OPT):
        pytest.skip()
        return

    # no need to clear memory paths
    return _get_storage_path(request, MEMORY)


@pytest.fixture
def local_path(request):
    if not is_opt_true(request, LOCAL_OPT):
        pytest.skip()
        return

    path = _get_storage_path(request, LOCAL)
    LocalProvider(path).clear()

    yield path

    # clear storage unless flagged otherwise
    if not is_opt_true(request, KEEP_STORAGE_OPT):
        LocalProvider(path).clear()


@pytest.fixture
def s3_path(request):
    if not is_opt_true(request, S3_OPT):
        pytest.skip()
        return

    path = _get_storage_path(request, S3)
    S3Provider(path).clear()

    yield path

    # clear storage unless flagged otherwise
    if not is_opt_true(request, KEEP_STORAGE_OPT):
        S3Provider(path).clear()


@pytest.fixture
def s3_vstream_path(request):
    if not is_opt_true(request, S3_OPT):
        pytest.skip()
        return

    path = f"{PYTEST_S3_PROVIDER_BASE_ROOT}vstream_test"
    yield path


@pytest.fixture(scope="session")
def gcs_creds():
    return os.environ.get(ENV_GOOGLE_APPLICATION_CREDENTIALS, None)


@pytest.fixture(scope="session")
def gdrive_creds():
    client_id = os.environ.get(ENV_GDRIVE_CLIENT_ID)
    client_secret = os.environ.get(ENV_GDRIVE_CLIENT_SECRET)
    refresh_token = os.environ.get(ENV_GDRIVE_REFRESH_TOKEN)
    creds = {
        "client_id": client_id,
        "client_secret": client_secret,
        "refresh_token": refresh_token,
    }
    return creds


@pytest.fixture
def gcs_path(request, gcs_creds):
    if not is_opt_true(request, GCS_OPT):
        pytest.skip()
        return

    path = _get_storage_path(request, GCS)
    GCSProvider(path, token=gcs_creds).clear()

    yield path

    # clear storage unless flagged otherwise
    if not is_opt_true(request, KEEP_STORAGE_OPT):
        GCSProvider(path, token=gcs_creds).clear()


@pytest.fixture
def gcs_vstream_path(request):
    if not is_opt_true(request, GCS_OPT):
        pytest.skip()
        return

    path = f"{PYTEST_GCS_PROVIDER_BASE_ROOT}vstream_test"
    yield path


@pytest.fixture
def gdrive_path(request, gdrive_creds):
    if not is_opt_true(request, GDRIVE_OPT):
        pytest.skip()
        return

    path = _get_storage_path(request, GDRIVE, with_current_test_name=False)
    GDriveProvider(path, token=gdrive_creds).clear()

    yield path

    if not is_opt_true(request, KEEP_STORAGE_OPT):
        GDriveProvider(path, token=gdrive_creds).clear()


@pytest.fixture
def hub_cloud_path(request, hub_cloud_dev_token):
    if not is_opt_true(request, HUB_CLOUD_OPT):
        pytest.skip()
        return

    path = _get_storage_path(request, HUB_CLOUD)
    storage_provider_from_hub_path(path, token=hub_cloud_dev_token).clear()

    yield path

    # clear storage unless flagged otherwise
    if not is_opt_true(request, KEEP_STORAGE_OPT):
        try:
            hub.delete(path, force=True, large_ok=True, token=hub_cloud_dev_token)
        except Exception:
            # TODO: investigate flakey `BadRequestException:
            # Invalid Request. One or more request parameters is incorrect.`
            # (on windows 3.8 only)
            pass


@pytest.fixture
def hub_cloud_vstream_path(request, hub_cloud_dev_token):
    if not is_opt_true(request, HUB_CLOUD_OPT):
        pytest.skip()
        return

    path = f"{PYTEST_HUB_CLOUD_PROVIDER_BASE_ROOT}vstream_test_dataset"

    yield path


@pytest.fixture
def cat_path():
    """Path to a cat image in the dummy data folder. Expected shape: (900, 900, 3)"""

    path = get_dummy_data_path("images")
    return os.path.join(path, "cat.jpeg")


@pytest.fixture
def flower_path():
    """Path to a flower image in the dummy data folder. Expected shape: (513, 464, 4)"""

    path = get_dummy_data_path("images")
    return os.path.join(path, "flower.png")


@pytest.fixture
def color_image_paths():
    base = get_dummy_data_path("images")
    paths = {
        "jpeg": os.path.join(base, "dog2.jpg"),
    }
    return paths


@pytest.fixture
def grayscale_image_paths():
    base = get_dummy_data_path("images")
    paths = {
        "jpeg": os.path.join(base, "hopper_gray.jpg"),
    }
    return paths


@pytest.fixture(scope="session")
def compressed_image_paths():
    paths = {
        "webp": "beach.webp",
        "gif": "boat.gif",
        "bmp": "car.bmp",
        "jpeg": ["cat.jpeg", "dog1.jpg", "dog2.jpg", "car.jpg"],
        "wmf": "crown.wmf",
        "dib": "dog.dib",
        "tiff": ["field.tiff", "field.tif"],
        "png": "flower.png",
        "ico": "sample_ico.ico",
        "jpeg2000": "sample_jpeg2000.jp2",
        "pcx": "sample_pcx.pcx",
        "ppm": "sample_ppm.ppm",
        "sgi": "sample_sgi.sgi",
        "tga": "sample_tga.tga",
        "xbm": "sample_xbm.xbm",
    }

    paths = {k: ([v] if isinstance(v, str) else v) for k, v in paths.items()}

    parent = get_dummy_data_path("images")
    for k in paths:
        paths[k] = [os.path.join(parent, p) for p in paths[k]]

    # Since we implement our own meta data reading for jpegs and pngs,
    # we test against images from PIL repo to cover all edge cases.
    pil_image_paths = _download_pil_test_images()
    paths["jpeg"] += pil_image_paths[".jpg"]
    paths["png"] += pil_image_paths[".png"]
    hub_test_images = _download_hub_test_images()
    paths["jpeg"] += hub_test_images
    yield paths


@pytest.fixture
def corrupt_image_paths():
    paths = {"jpeg": "corrupt_jpeg.jpeg", "png": "corrupt_png.png"}

    parent = get_dummy_data_path("images")
    for k in paths:
        paths[k] = os.path.join(parent, paths[k])

    return paths


@pytest.fixture
def audio_paths():
    if sys.platform.startswith("linux") and sys.version_info[:2] == (3, 6):  # FixMe
        pytest.skip()
        return
    paths = {"mp3": "samplemp3.mp3", "flac": "sampleflac.flac", "wav": "samplewav.wav"}

    parent = get_dummy_data_path("audio")
    for k in paths:
        paths[k] = os.path.join(parent, paths[k])

    return paths


@pytest.fixture
def video_paths():
    paths = {
        "mp4": ["samplemp4.mp4"],
        "mkv": ["samplemkv.mkv"],
        "avi": ["sampleavi.avi"],
    }

    parent = get_dummy_data_path("video")
    for k in paths:
        paths[k] = [os.path.join(parent, fname) for fname in paths[k]]
    paths["mp4"] += _download_hub_test_videos()

    return paths


@pytest.fixture
def point_cloud_paths():
    paths = {
        "las": "point_cloud.las",
    }

    parent = get_dummy_data_path("point_cloud")
    for k in paths:
        paths[k] = os.path.join(parent, paths[k])
    return paths


@pytest.fixture
def vstream_path(request):
    """Used with parametrize to use all video stream test datasets."""
    return request.getfixturevalue(request.param)


@pytest.fixture
def path(request):
    """Used with parametrize to get all dataset paths."""
    return request.getfixturevalue(request.param)


@pytest.fixture
def hub_token(request):
    """Used with parametrize to get hub_cloud_dev_token if hub-cloud option is True else None"""
    if is_opt_true(request, HUB_CLOUD_OPT):
        return request.getfixturevalue(request.param)
    return None



from importlib.util import find_spec


def pytorch_installed():
    return find_spec("torch") != None


def tensorflow_installed():
    return find_spec("tensorflow") != None


def tfds_installed():
    return find_spec("tensorflow_datasets") != None


def ray_installed():
    return find_spec("ray") != None



import hub
import pytest
from hub.tests.dataset_fixtures import enabled_non_gcs_gdrive_datasets


def test_text(memory_ds):
    ds = memory_ds
    ds.create_tensor("text", htype="text")
    items = ["abcd", "efgh", "0123456"]
    with ds:
        for x in items:
            ds.text.append(x)
        ds.text.extend(items)
    assert ds.text.shape == (6, 1)
    for i in range(6):
        assert ds.text[i].numpy()[0] == items[i % 3]


@enabled_non_gcs_gdrive_datasets
def test_text_transform(ds, scheduler="threaded"):
    ds.create_tensor("text", htype="text")

    @hub.compute
    def upload(some_str, ds):
        ds.text.append(some_str)
        return ds

    upload().eval(
        ["hi", "if ur reading this ur a nerd"], ds, num_workers=2, scheduler=scheduler
    )

    assert len(ds) == 2
    assert ds.text.data()["value"] == ["hi", "if ur reading this ur a nerd"]


@pytest.mark.parametrize(
    "args", [{}, {"sample_compression": "lz4"}, {"chunk_compression": "lz4"}]
)
def test_text_update(memory_ds, args):
    ds = memory_ds
    with ds:
        ds.create_tensor("x", htype="text", **args)
        for _ in range(10):
            ds.x.append("cat")
        assert ds.x[0].text() == "cat"

    for i in range(0, 10, 2):
        ds.x[i] = "flower"
    assert ds.x.data()["value"] == ["flower", "cat"] * 5
    assert ds.x.text() == ["flower", "cat"] * 5



"""
Helper function for exporting tensorflow dataset wrapper
"""
import warnings

from hub.util.exceptions import (
    ModuleNotInstalledException,
    SampleDecompressionError,
    CorruptedSampleError,
)
from hub.util.check_installation import tensorflow_installed


def dataset_to_tensorflow(dataset, tensors, tobytes):
    """Converts the dataset into a tensorflow compatible format"""
    if not tensorflow_installed():
        raise ModuleNotInstalledException(
            "'tensorflow' should be installed to convert the Dataset into tensorflow format"
        )

    import tensorflow as tf  # type: ignore
    from hub.integrations.tf.hubtensorflowdataset import HubTensorflowDataset  # type: ignore

    if not tensors:
        tensors = dataset.tensors

    if isinstance(tobytes, bool):
        tobytes = {k: tobytes for k in tensors}
    else:
        for k in tobytes:
            if k not in tensors:
                raise Exception(
                    f"Tensor {k} is not present in the list of provided tensors: {tensors}."
                )
        tobytes = {k: k in tobytes for k in tensors}

    def __iter__():
        for index in range(len(dataset)):
            sample = {}
            corrupt_sample_found = False
            for key in tensors:
                try:
                    value = dataset[key][index]
                    if tobytes[key]:
                        value = [value.tobytes()]
                    else:
                        value = value.numpy()
                    sample[key] = value
                except SampleDecompressionError:
                    warnings.warn(
                        f"Skipping corrupt {dataset[key].meta.sample_compression} sample."
                    )
                    corrupt_sample_found = True
            if not corrupt_sample_found:
                yield sample

    def generate_signature():
        signature = {}
        for key in tensors:
            tb = tobytes[key]
            dtype = dataset[key].meta.dtype
            if tb or dtype == "str":
                dtype = tf.string
            shape = (1,) if tb else dataset[key].shape[1:]
            signature[key] = tf.TensorSpec(shape=shape, dtype=dtype)
        return signature

    signature = generate_signature()
    return HubTensorflowDataset.from_generator(__iter__, output_signature=signature)



import click
import hub


@click.command()
@click.option(
    "--workspace", "-w", default=None, help="List your datasets in the workspace"
)
def list_datasets(workspace: str):
    """Get a list of datasets in the workspace from Platform."""
    try:
        res = hub.list(workspace)
        if res:
            click.echo("\n".join(res))
    except Exception as e:
        raise SystemExit(f"Unable to list datasets: {e}")



import pickle
from hub.experimental import dataloader

import hub
import numpy as np
import pytest
from hub.util.exceptions import TensorDoesNotExistError

from hub.util.remove_cache import get_base_storage
from hub.core.index.index import IndexEntry
from hub.tests.common import requires_torch, requires_linux
from hub.core.dataset import Dataset
from hub.core.storage import MemoryProvider, GCSProvider
from hub.constants import KB

from hub.tests.dataset_fixtures import enabled_non_gdrive_datasets

try:
    from torch.utils.data._utils.collate import default_collate
except ImportError:
    pass

# ensure tests have multiple chunks without a ton of data
PYTORCH_TESTS_MAX_CHUNK_SIZE = 5 * KB


def double(sample):
    return sample * 2


def to_tuple(sample):
    return sample["image"], sample["image2"]


def reorder_collate(batch):
    x = [((x["a"], x["b"]), x["c"]) for x in batch]
    return default_collate(x)


def dict_to_list(sample):
    return [sample["a"], sample["b"], sample["c"]]


def my_transform_collate(batch):
    x = [(c, a, b) for a, b, c in batch]
    return default_collate(x)


def index_transform(sample):
    return sample["index"], sample["xyz"]


@requires_torch
@enabled_non_gdrive_datasets
@requires_linux
def test_pytorch_small(ds):
    with ds:
        ds.create_tensor("image", max_chunk_size=PYTORCH_TESTS_MAX_CHUNK_SIZE)
        ds.image.extend(([i * np.ones((i + 1, i + 1)) for i in range(16)]))
        ds.commit()
        ds.create_tensor("image2", max_chunk_size=PYTORCH_TESTS_MAX_CHUNK_SIZE)
        ds.image2.extend(np.array([i * np.ones((12, 12)) for i in range(16)]))

    if isinstance(get_base_storage(ds.storage), (MemoryProvider, GCSProvider)):
        with pytest.raises(ValueError):
            dl = dataloader(ds)
        return
    dl = dataloader(ds).batch(1).pytorch(num_workers=2)

    assert len(dl.dataset) == 16

    for _ in range(2):
        for i, batch in enumerate(dl):
            np.testing.assert_array_equal(
                batch["image"].numpy(), i * np.ones((1, i + 1, i + 1))
            )
            np.testing.assert_array_equal(
                batch["image2"].numpy(), i * np.ones((1, 12, 12))
            )

    sub_ds = ds[5:]
    sub_dl = dataloader(sub_ds).pytorch(num_workers=0)

    for i, batch in enumerate(sub_dl):
        np.testing.assert_array_equal(
            batch["image"].numpy(), (5 + i) * np.ones((1, 6 + i, 6 + i))
        )
        np.testing.assert_array_equal(
            batch["image2"].numpy(), (5 + i) * np.ones((1, 12, 12))
        )

    sub_ds2 = ds[8:12]
    sub_dl2 = dataloader(sub_ds2).pytorch(num_workers=0)

    for _ in range(2):
        for i, batch in enumerate(sub_dl2):
            np.testing.assert_array_equal(
                batch["image"].numpy(), (8 + i) * np.ones((1, 9 + i, 9 + i))
            )
            np.testing.assert_array_equal(
                batch["image2"].numpy(), (8 + i) * np.ones((1, 12, 12))
            )

    sub_ds3 = ds[:5]
    sub_dl3 = dataloader(sub_ds3).pytorch(num_workers=0)

    for _ in range(2):
        for i, batch in enumerate(sub_dl3):
            np.testing.assert_array_equal(
                batch["image"].numpy(), (i) * np.ones((1, i + 1, i + 1))
            )
            np.testing.assert_array_equal(
                batch["image2"].numpy(), (i) * np.ones((1, 12, 12))
            )


@requires_torch
@enabled_non_gdrive_datasets
@requires_linux
def test_pytorch_transform(ds):
    with ds:
        ds.create_tensor("image", max_chunk_size=PYTORCH_TESTS_MAX_CHUNK_SIZE)
        ds.image.extend(([i * np.ones((i + 1, i + 1)) for i in range(16)]))
        ds.checkout("alt", create=True)
        ds.create_tensor("image2", max_chunk_size=PYTORCH_TESTS_MAX_CHUNK_SIZE)
        ds.image2.extend(np.array([i * np.ones((12, 12)) for i in range(16)]))

    if isinstance(get_base_storage(ds.storage), (MemoryProvider, GCSProvider)):
        with pytest.raises(ValueError):
            dl = dataloader(ds)
        return

    dl = dataloader(ds).batch(1).transform(to_tuple).pytorch(num_workers=2)

    for _ in range(2):
        for i, batch in enumerate(dl):
            actual_image = batch[0].numpy()
            expected_image = i * np.ones((1, i + 1, i + 1))
            actual_image2 = batch[1].numpy()
            expected_image2 = i * np.ones((1, 12, 12))
            np.testing.assert_array_equal(actual_image, expected_image)
            np.testing.assert_array_equal(actual_image2, expected_image2)


@requires_torch
@enabled_non_gdrive_datasets
@requires_linux
def test_pytorch_transform_dict(ds):
    with ds:
        ds.create_tensor("image", max_chunk_size=PYTORCH_TESTS_MAX_CHUNK_SIZE)
        ds.image.extend(([i * np.ones((i + 1, i + 1)) for i in range(16)]))
        ds.create_tensor("image2", max_chunk_size=PYTORCH_TESTS_MAX_CHUNK_SIZE)
        ds.image2.extend(np.array([i * np.ones((12, 12)) for i in range(16)]))
        ds.create_tensor("image3", max_chunk_size=PYTORCH_TESTS_MAX_CHUNK_SIZE)
        ds.image3.extend(np.array([i * np.ones((12, 12)) for i in range(16)]))

    if isinstance(get_base_storage(ds.storage), (MemoryProvider, GCSProvider)):
        with pytest.raises(ValueError):
            dl = dataloader(ds)
        return

    dl = dataloader(ds).transform({"image": double, "image2": None}).pytorch()

    assert len(dl.dataset) == 16

    for _ in range(2):
        for i, batch in enumerate(dl):
            assert set(batch.keys()) == {"image", "image2"}
            np.testing.assert_array_equal(
                batch["image"].numpy(), 2 * i * np.ones((1, i + 1, i + 1))
            )
            np.testing.assert_array_equal(
                batch["image2"].numpy(), i * np.ones((1, 12, 12))
            )

    for _ in range(2):
        for i, (image, image2) in enumerate(dl):
            np.testing.assert_array_equal(
                image.numpy(), 2 * i * np.ones((1, i + 1, i + 1))
            )
            np.testing.assert_array_equal(image2.numpy(), i * np.ones((1, 12, 12)))


@requires_torch
@enabled_non_gdrive_datasets
@requires_linux
def test_pytorch_with_compression(ds: Dataset):
    # TODO: chunk-wise compression for labels (right now they are uncompressed)
    with ds:
        images = ds.create_tensor(
            "images",
            htype="image",
            sample_compression="png",
            max_chunk_size=PYTORCH_TESTS_MAX_CHUNK_SIZE,
        )
        labels = ds.create_tensor(
            "labels", htype="class_label", max_chunk_size=PYTORCH_TESTS_MAX_CHUNK_SIZE
        )

        assert images.meta.sample_compression == "png"

        images.extend(np.ones((16, 12, 12, 3), dtype="uint8"))
        labels.extend(np.ones((16, 1), dtype="uint32"))

    if isinstance(get_base_storage(ds.storage), (MemoryProvider, GCSProvider)):
        with pytest.raises(ValueError):
            dl = dataloader(ds)
        return

    dl = dataloader(ds).pytorch(num_workers=0)

    for _ in range(2):
        for batch in dl:
            X = batch["images"].numpy()
            T = batch["labels"].numpy()
            assert X.shape == (1, 12, 12, 3)
            assert T.shape == (1, 1)


@requires_torch
@enabled_non_gdrive_datasets
@requires_linux
def test_custom_tensor_order(ds):
    with ds:
        tensors = ["a", "b", "c", "d"]
        for t in tensors:
            ds.create_tensor(t, max_chunk_size=PYTORCH_TESTS_MAX_CHUNK_SIZE)
            ds[t].extend(np.random.random((3, 4, 5)))

    if isinstance(get_base_storage(ds.storage), (MemoryProvider, GCSProvider)):
        with pytest.raises(ValueError):
            dl = dataloader(ds)
        return

    with pytest.raises(TensorDoesNotExistError):
        dl = dataloader(ds).pytorch(tensors=["c", "d", "e"])

    dl = dataloader(ds).pytorch(tensors=["c", "d", "a"], return_index=False)

    for i, batch in enumerate(dl):
        c1, d1, a1 = batch
        a2 = batch["a"]
        c2 = batch["c"]
        d2 = batch["d"]
        assert "b" not in batch
        np.testing.assert_array_equal(a1, a2)
        np.testing.assert_array_equal(c1, c2)
        np.testing.assert_array_equal(d1, d2)
        np.testing.assert_array_equal(a1[0], ds.a.numpy()[i])
        np.testing.assert_array_equal(c1[0], ds.c.numpy()[i])
        np.testing.assert_array_equal(d1[0], ds.d.numpy()[i])
        batch = pickle.loads(pickle.dumps(batch))
        c1, d1, a1 = batch
        a2 = batch["a"]
        c2 = batch["c"]
        d2 = batch["d"]
        np.testing.assert_array_equal(a1, a2)
        np.testing.assert_array_equal(c1, c2)
        np.testing.assert_array_equal(d1, d2)
        np.testing.assert_array_equal(a1[0], ds.a.numpy()[i])
        np.testing.assert_array_equal(c1[0], ds.c.numpy()[i])
        np.testing.assert_array_equal(d1[0], ds.d.numpy()[i])


@requires_torch
@requires_linux
def test_readonly_with_two_workers(local_ds):
    local_ds.create_tensor("images", max_chunk_size=PYTORCH_TESTS_MAX_CHUNK_SIZE)
    local_ds.create_tensor("labels", max_chunk_size=PYTORCH_TESTS_MAX_CHUNK_SIZE)
    local_ds.images.extend(np.ones((10, 12, 12)))
    local_ds.labels.extend(np.ones(10))

    base_storage = get_base_storage(local_ds.storage)
    base_storage.flush()
    base_storage.enable_readonly()
    ds = Dataset(storage=local_ds.storage, read_only=True, verbose=False)

    ptds = dataloader(ds).pytorch(num_workers=2)
    # no need to check input, only care that readonly works
    for _ in ptds:
        pass


@pytest.mark.xfail(raises=NotImplementedError, strict=True)
def test_corrupt_dataset():
    raise NotImplementedError


@pytest.mark.xfail(raises=NotImplementedError, strict=True)
def test_pytorch_local_cache():
    raise NotImplementedError


@requires_torch
@requires_linux
def test_groups(local_ds, compressed_image_paths):
    img1 = hub.read(compressed_image_paths["jpeg"][0])
    img2 = hub.read(compressed_image_paths["png"][0])
    with local_ds:
        local_ds.create_tensor(
            "images/jpegs/cats", htype="image", sample_compression="jpeg"
        )
        local_ds.create_tensor(
            "images/pngs/flowers", htype="image", sample_compression="png"
        )
        for _ in range(10):
            local_ds.images.jpegs.cats.append(img1)
            local_ds.images.pngs.flowers.append(img2)

    another_ds = hub.dataset(local_ds.path)
    dl = dataloader(another_ds).pytorch(return_index=False)
    for i, (cat, flower) in enumerate(dl):
        assert cat[0].shape == another_ds.images.jpegs.cats[i].numpy().shape
        assert flower[0].shape == another_ds.images.pngs.flowers[i].numpy().shape


@requires_torch
@requires_linux
def test_string_tensors(local_ds):
    with local_ds:
        local_ds.create_tensor("strings", htype="text")
        local_ds.strings.extend([f"string{idx}" for idx in range(5)])

    ptds = dataloader(local_ds).pytorch()
    for idx, batch in enumerate(ptds):
        np.testing.assert_array_equal(batch["strings"], f"string{idx}")


@pytest.mark.xfail(raises=NotImplementedError, strict=True)
def test_pytorch_large():
    raise NotImplementedError


@requires_torch
@requires_linux
@pytest.mark.parametrize(
    "index",
    [
        slice(2, 7),
        slice(3, 10, 2),
        slice(None, 10),
        # slice(None, None, -1),
        # slice(None, None, -2),
        [2, 3, 4],
        [2, 4, 6, 8],
        [2, 2, 4, 4, 6, 6, 7, 7, 8, 8, 9, 9, 9],
        [4, 3, 2, 1],
    ],
)
def test_pytorch_view(local_ds, index):
    arr_list_1 = [np.random.randn(15, 15, i) for i in range(10)]
    arr_list_2 = [np.random.randn(40, 15, 4, i) for i in range(10)]
    label_list = list(range(10))

    with local_ds as ds:
        ds.create_tensor("img1")
        ds.create_tensor("img2")
        ds.create_tensor("label")
        ds.img1.extend(arr_list_1)
        ds.img2.extend(arr_list_2)
        ds.label.extend(label_list)

    ptds = dataloader(local_ds[index]).pytorch()
    idxs = list(IndexEntry(index).indices(len(local_ds)))
    for idx, batch in enumerate(ptds):
        idx = idxs[idx]
        np.testing.assert_array_equal(batch["img1"][0], arr_list_1[idx])
        np.testing.assert_array_equal(batch["img2"][0], arr_list_2[idx])
        np.testing.assert_array_equal(batch["label"][0], idx)


@requires_torch
@requires_linux
def test_pytorch_collate(local_ds):
    with local_ds:
        local_ds.create_tensor("a")
        local_ds.create_tensor("b")
        local_ds.create_tensor("c")
        for _ in range(100):
            local_ds.a.append(0)
            local_ds.b.append(1)
            local_ds.c.append(2)

    ptds = (
        dataloader(local_ds)
        .batch(4)
        .pytorch(
            collate_fn=reorder_collate,
        )
    )
    for batch in ptds:
        assert len(batch) == 2
        assert len(batch[0]) == 2
        np.testing.assert_array_equal(batch[0][0], np.array([0, 0, 0, 0]).reshape(4, 1))
        np.testing.assert_array_equal(batch[0][1], np.array([1, 1, 1, 1]).reshape(4, 1))
        np.testing.assert_array_equal(batch[1], np.array([2, 2, 2, 2]).reshape(4, 1))


@requires_torch
@requires_linux
def test_pytorch_transform_collate(local_ds):
    with local_ds:
        local_ds.create_tensor("a")
        local_ds.create_tensor("b")
        local_ds.create_tensor("c")
        for _ in range(100):
            local_ds.a.append(0 * np.ones((300, 300)))
            local_ds.b.append(1 * np.ones((300, 300)))
            local_ds.c.append(2 * np.ones((300, 300)))

    ptds = (
        dataloader(local_ds)
        .batch(4)
        .pytorch(
            collate_fn=my_transform_collate,
        )
        .transform(dict_to_list)
    )
    for batch in ptds:
        assert len(batch) == 3
        for i in range(2):
            assert len(batch[i]) == 4
        np.testing.assert_array_equal(batch[0], 2 * np.ones((4, 300, 300)))
        np.testing.assert_array_equal(batch[1], 0 * np.ones((4, 300, 300)))
        np.testing.assert_array_equal(batch[2], 1 * np.ones((4, 300, 300)))


@pytest.mark.xfail(raises=NotImplementedError, strict=True)
def test_pytorch_ddp():
    raise NotImplementedError


@pytest.mark.xfail(raises=NotImplementedError, strict=True)
def test_pytorch_tobytes():
    raise NotImplementedError


@requires_torch
@requires_linux
def test_rename(local_ds):
    with local_ds as ds:
        ds.create_tensor("abc")
        ds.create_tensor("blue/green")
        ds.abc.append([1, 2, 3])
        ds.rename_tensor("abc", "xyz")
        ds.rename_group("blue", "red")
        ds["red/green"].append([1, 2, 3, 4])
    loader = dataloader(ds).pytorch(return_index=False)
    for sample in loader:
        assert set(sample.keys()) == {"xyz", "red/green"}
        np.testing.assert_array_equal(np.array(sample["xyz"]), np.array([[1, 2, 3]]))
        np.testing.assert_array_equal(
            np.array(sample["red/green"]), np.array([[1, 2, 3, 4]])
        )


@requires_torch
@requires_linux
def test_expiration_date_casting_to_string():
    ds = hub.dataset("hub://activeloop/cifar100-train")[0:10:2]
    loader = dataloader(ds).pytorch(return_index=False)
    for _ in loader:
        pass


@requires_torch
@requires_linux
@pytest.mark.parametrize("num_workers", [0, 2])
def test_indexes(local_ds, num_workers):
    with local_ds as ds:
        ds.create_tensor("xyz")
        for i in range(8):
            ds.xyz.append(i * np.ones((2, 2)))

    ptds = (
        dataloader(local_ds)
        .batch(4)
        .pytorch(num_workers=num_workers, return_index=True)
    )

    for batch in ptds:
        assert batch.keys() == {"xyz", "index"}
        for i in range(len(batch)):
            np.testing.assert_array_equal(batch["index"][i], batch["xyz"][i][0, 0])


@requires_torch
@requires_linux
@pytest.mark.parametrize("num_workers", [0, 2])
def test_indexes_transform(local_ds, num_workers):
    with local_ds as ds:
        ds.create_tensor("xyz")
        for i in range(8):
            ds.xyz.append(i * np.ones((2, 2)))

    ptds = (
        dataloader(local_ds)
        .batch(4)
        .transform(index_transform)
        .pytorch(num_workers=num_workers, return_index=True)
    )

    for batch in ptds:
        assert len(batch) == 2
        assert len(batch[0]) == 4
        assert len(batch[1]) == 4

        for i in range(4):
            np.testing.assert_array_equal(batch[0][i], batch[1][i][0, 0])


@requires_torch
@requires_linux
@pytest.mark.parametrize("num_workers", [0, 2])
def test_indexes_transform_dict(local_ds, num_workers):
    with local_ds as ds:
        ds.create_tensor("xyz")
        for i in range(8):
            ds.xyz.append(i * np.ones((2, 2)))

    ptds = (
        dataloader(local_ds)
        .batch(4)
        .transform({"xyz": double, "index": None})
        .pytorch(num_workers=num_workers, return_index=True)
    )

    for batch in ptds:
        assert batch.keys() == {"xyz", "index"}
        for i in range(len(batch)):
            np.testing.assert_array_equal(2 * batch["index"][i], batch["xyz"][i][0, 0])

    ptds = (
        dataloader(local_ds)
        .batch(4)
        .transform({"xyz": double})
        .pytorch(num_workers=num_workers, return_index=True)
    )

    for batch in ptds:
        assert batch.keys() == {"xyz"}


@requires_torch
@requires_linux
@pytest.mark.parametrize("num_workers", [0, 2])
def test_indexes_tensors(local_ds, num_workers):
    with local_ds as ds:
        ds.create_tensor("xyz")
        for i in range(8):
            ds.xyz.append(i * np.ones((2, 2)))

    with pytest.raises(ValueError):
        ptds = (
            dataloader(local_ds)
            .batch(4)
            .pytorch(
                num_workers=num_workers, return_index=True, tensors=["xyz", "index"]
            )
        )

    ptds = (
        dataloader(local_ds)
        .batch(4)
        .pytorch(num_workers=num_workers, return_index=True, tensors=["xyz"])
    )

    for batch in ptds:
        assert batch.keys() == {"xyz", "index"}


@requires_linux
def test_uneven_iteration(local_ds):
    with local_ds as ds:
        ds.create_tensor("x")
        ds.create_tensor("y")
        ds.x.extend(list(range(5)))
        ds.y.extend(list(range(10)))
    ptds = dataloader(ds).pytorch()
    for i, batch in enumerate(ptds):
        x, y = np.array(batch["x"][0]), np.array(batch["y"][0])
        np.testing.assert_equal(x, i)
        np.testing.assert_equal(y, i)



from typing import Any, Dict, List, Optional, Tuple, Union
import numpy as np
from numpy import ndarray
import json
import base64
from hub.core.sample import Sample  # type: ignore

Schema = Any


scalars = ["int", "float", "bool", "str", "list", "dict", "ndarray", "Sample"]
types = ["Any", "Dict", "List", "Optional", "Union"]


def _norm_type(typ: str):
    typ = typ.replace("typing.", "")
    replacements = {
        "numpy.ndarray": "ndarray",
        "np.ndarray": "ndarray",
        "hub.core.sample.Sample": "Sample",
        "hub.Sample": "Sample",
    }
    return replacements.get(typ, typ)


def _parse_schema(schema: Union[str, Schema]) -> Tuple[str, List[str]]:
    if getattr(schema, "__module__", None) == "typing":
        schema = str(schema)
        validate = False
    else:
        validate = True

    if schema in scalars:
        return schema, []

    if "[" not in schema:
        return _norm_type(schema), []

    typ, param_string = schema.split("[", 1)
    typ = _norm_type(typ)
    assert param_string[-1] == "]"
    params = []
    buff = ""
    level = 0
    for c in param_string:
        if c == "[":
            level += 1
            buff += c
        elif c == "]":
            if level == 0:
                if buff:
                    params.append(buff)
                if validate:
                    _validate_schema(typ, params)
                return typ, params
            else:
                buff += c
                level -= 1
        elif c == ",":
            if level == 0:
                params.append(buff)
                buff = ""
            else:
                buff += c
        elif c == " ":
            continue
        else:
            buff += c
    raise InvalidJsonSchemaException()


class InvalidJsonSchemaException(Exception):
    pass


class ArgumentMismatchException(InvalidJsonSchemaException):
    def __init__(self, typ: str, actual: int, expected: int, exact: bool = False):
        assert actual != expected
        gt = actual > expected
        super(ArgumentMismatchException, self).__init__(
            f"Too {'many' if gt else 'few'} parameters for {typ};"
            + f" actual {actual},expected {'exatcly' if exact else ('at most' if gt else 'at least')} {expected}."
        )


def _validate_schema(typ: str, params: List[str]) -> Tuple[str, List[str]]:
    if typ in scalars:
        return typ, params

    if typ not in types:
        raise InvalidJsonSchemaException(f"Unsupported type: {typ}")

    def _err(expected_num_params: int, exact: bool = False):
        raise ArgumentMismatchException(typ, len(params), expected_num_params, exact)

    if typ == "Any":
        if params:
            _err(0)
    elif typ == "Optional":
        if len(params) > 1:
            _err(1)
    elif typ == "Union":
        if len(params) == 0:
            _err(1)
    elif typ == "List":
        if len(params) > 1:
            _err(1)
    elif typ == "Dict":
        if len(params) not in (0, 2):
            _err(2, True)
    return typ, params


def _validate_any(obj: Any, params: List[str]):
    assert not params
    return True


def _validate_union(obj: Any, params: List[str]):
    for schema in params:
        if _validate_object(obj, schema):
            return True
    return False


def _validate_optional(obj: Any, params: List[str]) -> bool:
    assert len(params) <= 1
    if obj is None:
        return True
    if params:
        return _validate_object(obj, params[0])
    return True


def _validate_list(obj: Any, params: List[str]) -> bool:
    assert len(params) <= 1
    if not isinstance(obj, (list, tuple)):
        return False
    if params:
        for item in obj:
            if not _validate_object(item, params[0]):
                return False
    return True


def _validate_dict(obj: Any, params: List[str]) -> bool:
    assert len(params) in (0, 2)
    if not isinstance(obj, dict):
        return False
    if params:
        assert params[0] in (
            "str",
            "Any",
        ), "Only string keys are allowed for json dicts."
        for v in obj.values():
            if not _validate_object(v, params[1]):
                return False
    return True


def _validate_nonetype(obj: Any, params: List[str]) -> bool:
    assert not params
    return obj is None


def _validate_object(obj: Any, schema: Union[str, Schema]) -> bool:
    typ, params = _parse_schema(schema)
    if typ in scalars:
        return isinstance(obj, eval(typ))
    return globals()[f"_validate_{typ.lower()}"](obj, params)


class JsonValidationError(Exception):
    pass


def validate_json_object(obj: Any, schema: Union[str, Schema]) -> None:
    if obj and not _validate_object(obj, schema):
        raise JsonValidationError()


def validate_json_schema(schema: str):
    _parse_schema(schema)


class HubJsonEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, ndarray):
            return {
                "_hub_custom_type": "ndarray",
                "data": base64.b64encode(obj.tobytes()).decode(),
                "shape": obj.shape,
                "dtype": obj.dtype.name,
            }
        elif isinstance(obj, Sample):
            if obj.compression:
                return {
                    "_hub_custom_type": "Sample",
                    "data": base64.b64encode(obj.buffer).decode(),
                    "compression": obj.compression,
                }
            else:
                return self.default(obj.array)
        return obj


class HubJsonDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(self, object_hook=self.object_hook, *args, **kwargs)

    def object_hook(self, obj):
        hub_custom_type = obj.get("_hub_custom_type")
        if hub_custom_type == "ndarray":
            return np.frombuffer(
                base64.b64decode(obj["data"]), dtype=obj["dtype"]
            ).reshape(obj["shape"])
        elif hub_custom_type == "Sample":
            return Sample(
                buffer=base64.b64decode(obj["data"]), compression=obj["compression"]
            )
        return obj



import pytest
import numpy as np

from hub.constants import MB

compressions_paremetrized = pytest.mark.parametrize(
    "compression",
    [
        {"sample_compression": None},
        {"sample_compression": "png"},
        {"chunk_compression": "png"},
        {"sample_compression": "lz4"},
        {"chunk_compression": "lz4"},
    ],
)


def test_simple(memory_ds):
    with memory_ds:
        memory_ds.create_tensor("abc", max_chunk_size=2 * MB, tiling_threshold=2 * MB)
        memory_ds.abc.extend(np.ones((3, 253, 501, 5)))
    np.testing.assert_array_equal(memory_ds.abc.numpy(), np.ones((3, 253, 501, 5)))
    memory_ds.commit()
    np.testing.assert_array_equal(memory_ds.abc.numpy(), np.ones((3, 253, 501, 5)))


@compressions_paremetrized
def test_mixed_small_large(local_ds_generator, compression):
    ds = local_ds_generator()
    arr1 = np.random.randint(0, 255, (3003, 2001, 3)).astype(np.uint8)
    arr2 = np.random.randint(0, 255, (500, 500, 3)).astype(np.uint8)
    arr3 = np.random.randint(0, 255, (250, 250, 3)).astype(np.uint8)

    idxs = [
        (slice(73, 117), slice(9, 17)),
        4,
        -1,
        slice(
            10,
        ),
        slice(20, 37),
    ]

    with ds:
        ds.create_tensor(
            "abc", max_chunk_size=2**21, tiling_threshold=2**20, **compression
        )
        for i in range(10):
            if i % 5 == 0:
                ds.abc.append(arr1)
            else:
                ds.abc.append(arr2)

    for i in range(10):
        if i % 5 == 0:
            np.testing.assert_array_equal(ds.abc[i].numpy(), arr1)
            for idx in idxs:
                np.testing.assert_array_equal(ds.abc[i][idx].numpy(), arr1[idx])
        else:
            np.testing.assert_array_equal(ds.abc[i].numpy(), arr2)
            for idx in idxs:
                np.testing.assert_array_equal(ds.abc[i][idx].numpy(), arr2[idx])

    ds = local_ds_generator()

    for i in range(10):
        if i % 5 == 0:
            np.testing.assert_array_equal(ds.abc[i].numpy(), arr1)
            for idx in idxs:
                np.testing.assert_array_equal(ds.abc[i][idx].numpy(), arr1[idx])
        else:
            np.testing.assert_array_equal(ds.abc[i].numpy(), arr2)
            for idx in idxs:
                np.testing.assert_array_equal(ds.abc[i][idx].numpy(), arr2[idx])

    with ds:
        ds.abc.extend([arr3] * 3)

    for i in range(13):
        if i >= 10:
            np.testing.assert_array_equal(ds.abc[i].numpy(), arr3)
        elif i % 5 == 0:
            np.testing.assert_array_equal(ds.abc[i].numpy(), arr1)
        else:
            np.testing.assert_array_equal(ds.abc[i].numpy(), arr2)

    ds = local_ds_generator()
    for i in range(13):
        if i >= 10:
            np.testing.assert_array_equal(ds.abc[i].numpy(), arr3)
        elif i % 5 == 0:
            np.testing.assert_array_equal(ds.abc[i].numpy(), arr1)
        else:
            np.testing.assert_array_equal(ds.abc[i].numpy(), arr2)


@compressions_paremetrized
def test_updates(memory_ds, compression):
    arr1 = np.random.randint(0, 255, (3003, 2001, 3)).astype(np.uint8)
    arr2 = np.random.randint(0, 255, (500, 500, 3)).astype(np.uint8)
    arr3 = np.random.randint(0, 255, (2503, 2501, 3)).astype(np.uint8)
    arr4 = np.random.randint(0, 255, (250, 250, 3)).astype(np.uint8)

    update_idx = (slice(73, 117), slice(1765, 1901))

    arr5 = arr1 * 2
    arr6 = arr5[update_idx]
    arr6 += 1

    with memory_ds:
        memory_ds.create_tensor(
            "abc", max_chunk_size=2**21, tiling_threshold=2**20, **compression
        )
        for i in range(10):
            if i % 5 == 0:
                memory_ds.abc.append(arr1)
            else:
                memory_ds.abc.append(arr2)
            len(memory_ds)
    with memory_ds:
        for i in range(10):
            if i % 5 == 0:
                memory_ds.abc[i] = arr1 * 2
                memory_ds.abc[i][update_idx] = arr6
            else:
                memory_ds.abc[i] = arr3 if i % 2 == 0 else arr4

    for i in range(10):
        if i % 5 == 0:
            np.testing.assert_array_equal(memory_ds.abc[i].numpy(), arr5)
        elif i % 2 == 0:
            np.testing.assert_array_equal(memory_ds.abc[i].numpy(), arr3)
        else:
            np.testing.assert_array_equal(memory_ds.abc[i].numpy(), arr4)

    # update tiled sample with small sample
    arr7 = np.random.randint(0, 255, (3, 2, 3)).astype(np.uint8)
    memory_ds.abc[0] = arr7
    np.testing.assert_array_equal(memory_ds.abc[0].numpy(), arr7)


def test_cachable_overflow(memory_ds):
    ds = memory_ds
    with ds:
        ds.create_tensor("x")
        ds.create_tensor("y")
        ds.x.extend(np.ones((3, 4000, 3000)))
        ds.y.extend(np.ones((3, 4000, 3000)))
    assert len(ds) == 3
    assert len(ds.x) == 3
    assert len(ds.y) == 3


@compressions_paremetrized
def test_empty_array(memory_ds, compression):
    ds = memory_ds
    arr_list = [
        np.random.randint(0, 255, (3894, 0, 3), dtype=np.uint8),
        np.random.randint(0, 255, (1089, 1027, 3), dtype=np.uint8),
    ]
    with ds:
        ds.create_tensor(
            "x", **compression, max_chunk_size=1 * MB, tiling_threshold=1 * MB
        )
        ds.x.extend(arr_list)
    assert len(ds) == 2
    assert len(ds.x) == 2

    for i in range(2):
        np.testing.assert_array_equal(ds.x[i].numpy(), arr_list[i])


@compressions_paremetrized
def test_no_tiling(memory_ds, compression):
    ds = memory_ds
    arr_list = [
        np.random.randint(0, 255, (3894, 4279, 3), dtype=np.uint8),
        np.random.randint(0, 255, (1089, 1027, 3), dtype=np.uint8),
    ]
    with ds:
        ds.create_tensor("x", **compression, max_chunk_size=1 * MB, tiling_threshold=-1)
        ds.x.extend(arr_list)
    assert len(ds) == 2
    assert len(ds.x) == 2
    assert ds.x.chunk_engine.num_chunks == 2


def test_chunk_sizes(memory_ds):
    ds = memory_ds
    with ds:
        ds.create_tensor("image", htype="image", sample_compression="png")
        sample = np.random.randint(0, 256, size=(5000, 5000, 3), dtype=np.uint8)
        ds.image.append(sample)
    np.testing.assert_array_equal(ds.image[0].numpy(), sample)
    num_chunks = 0
    for k in ds.storage:
        if k.startswith("image/chunks/"):
            chunk = ds.storage[k].tobytes()
            num_chunks += 1
            assert 16 * MB < len(chunk) < 20 * MB
    assert num_chunks == 4



from hub.core.meta.encode.byte_positions import BytePositionsEncoder
from hub.core.storage.hub_memory_object import HubMemoryObject
from hub.core.serialize import (
    serialize_sequence_or_creds_encoder,
    deserialize_sequence_or_creds_encoder,
)


class SequenceEncoder(BytePositionsEncoder, HubMemoryObject):
    @classmethod
    def frombuffer(cls, buffer: bytes):
        instance = cls()
        if not buffer:
            return instance
        version, ids = deserialize_sequence_or_creds_encoder(buffer, "seq")
        if ids.nbytes:
            instance._encoded = ids
        instance.version = version
        instance.is_dirty = False
        return instance

    def tobytes(self) -> memoryview:
        return memoryview(
            serialize_sequence_or_creds_encoder(self.version, self._encoded)
        )

    def pop(self, index):
        self.is_dirty = True
        super().pop(index)



import glob
import os, random
from collections import Counter
from typing import Tuple
import shutil
from hub.util.exceptions import AutoCompressionError
from hub.compression import IMAGE_COMPRESSION_EXTENSIONS


def get_most_common_extension(
    local_path: str, allowed_extensions: Tuple = tuple(IMAGE_COMPRESSION_EXTENSIONS)
):
    """Determines the most frequently used extension in a directory of files.

    Args:
        local_path (str): Directory to scan.
        allowed_extensions (Tuple): File extensions considered for scanning.

    Raises:
        AutoCompressionError: If no files with allowed extensions are found.

    Returns:
        compression (str): Most common extension under the provided path.
    """

    # Return file extension if path is not a directory
    if not os.path.isdir(local_path):
        file_extension = os.path.splitext(local_path)[1].split(".")[1]
        if file_extension is None:
            return None
        else:
            return file_extension

    g = glob.glob(os.path.join(local_path, "**"), recursive=True)

    file_names = [name for name in g if name.endswith(allowed_extensions)]

    if not file_names:
        raise AutoCompressionError(local_path)

    if len(file_names) > 100:
        file_names = random.sample(file_names, 100)

    extension_list = [
        os.path.splitext(file_names[file])[1] for file in range(len(file_names))
    ]

    most_common_extension = [
        extension
        for extension, extension_count in Counter(extension_list).most_common(3)
    ]
    compression = most_common_extension[0].split(".")[1]

    return compression


def ingestion_summary(local_path: str, skipped_files: list):
    """Generate post ingesiton summary in a tree structure.

    Args:
        local_path (str): Root directory of dataset.
        skipped_files (list): List of files skipped during ingestion.
    """
    columns, lines = shutil.get_terminal_size()

    mid = int(columns / 2)
    print("\n")

    if not skipped_files:
        print("Ingestion Complete. No files were skipped.")
        print("\n")
        return

    at_root = True
    for root, dirs, files in os.walk(local_path):
        files = [f for f in files if not f[0] == "."]
        dirs[:] = [d for d in dirs if not d[0] == "."]
        dirs.sort()

        level = root.replace(local_path, "").count(os.sep)
        indent = " " * 6 * (level)
        if at_root == True:
            print(
                "{}{}/    ".format(
                    indent,
                    os.path.basename(root),
                )
            )
            at_root = False
        else:
            print(
                "{}{}/    ".format(
                    indent,
                    os.path.basename(root),
                )
            )

        subindent = " " * 6 * (level + 1)
        for f in files:
            if f in skipped_files:
                print("{}[Skipped]  {}".format(subindent, f))



from hub.core.link_creds import LinkCreds
from hub.core.lock import Lock
from hub.core.storage.lru_cache import LRUCache
from hub.util.keys import (
    get_dataset_linked_creds_key,
    get_dataset_linked_creds_lock_key,
)
from hub.util.remove_cache import get_base_storage
from typing import Optional, Tuple


def merge_link_creds(
    old_link_creds: LinkCreds,
    current_link_creds: LinkCreds,
    replaced_index: Optional[int] = None,
    managed_info: Optional[Tuple] = None,
):
    num_common_keys = 0
    if replaced_index is not None:
        new_key = current_link_creds.creds_keys[replaced_index]
    for i, (key1, key2) in enumerate(
        zip(old_link_creds.creds_keys, current_link_creds.creds_keys)
    ):
        if key1 == key2 or i == replaced_index:
            num_common_keys += 1
        else:
            break
    new_keys = current_link_creds.creds_keys[num_common_keys:]
    current_link_creds.creds_keys = old_link_creds.creds_keys
    current_link_creds.creds_mapping = old_link_creds.creds_mapping
    current_link_creds.managed_creds_keys = old_link_creds.managed_creds_keys
    current_link_creds.used_creds_keys = old_link_creds.used_creds_keys.union(
        current_link_creds.used_creds_keys
    )
    for key in new_keys:
        if key not in current_link_creds.creds_mapping:
            managed = key in current_link_creds.managed_creds_keys
            current_link_creds.add_creds_key(key, managed)
    if replaced_index is not None:
        replaced_key = current_link_creds.creds_keys[replaced_index]
        current_link_creds.replace_creds(replaced_key, new_key)
    if managed_info is not None:
        is_managed = managed_info[0]
        managed_index = managed_info[1]
        key = current_link_creds.creds_keys[managed_index]
        if is_managed:
            current_link_creds.managed_creds_keys.add(key)
        else:
            current_link_creds.managed_creds_keys.discard(key)
    return current_link_creds


def save_link_creds(
    current_link_creds: LinkCreds,
    storage: LRUCache,
    replaced_index: Optional[int] = None,
    managed_info: Optional[Tuple] = None,
):
    """Saves the linked creds info to storage."""
    storage = get_base_storage(storage)
    lock = Lock(storage, get_dataset_linked_creds_lock_key())
    lock.acquire(timeout=10, force=True)
    key = get_dataset_linked_creds_key()
    try:
        data_bytes = storage[key]
    except KeyError:
        data_bytes = None

    if data_bytes is not None:
        old_link_creds = LinkCreds.frombuffer(data_bytes)
        new_link_creds = merge_link_creds(
            old_link_creds, current_link_creds, replaced_index, managed_info
        )
    else:
        new_link_creds = current_link_creds

    storage[key] = new_link_creds.tobytes()
    lock.release()


def get_path_creds_key(sample):
    if sample is None:
        return None, None
    return sample.path, sample.creds_key



import os

TOKEN_FILE_PATH = os.path.expanduser("~/.activeloop/token")
HUB_PYPI_VERSION_PATH = os.path.expanduser("~/.activeloop/pypi_version.json")
REPORTING_CONFIG_FILE_PATH = os.path.expanduser("~/.activeloop/reporting_config.json")

HUB_REST_ENDPOINT = "https://app.activeloop.ai"
HUB_REST_ENDPOINT_STAGING = "https://app-staging.activeloop.dev"
HUB_REST_ENDPOINT_DEV = "https://app-dev.activeloop.dev"
HUB_REST_ENDPOINT_LOCAL = "http://localhost:7777"
USE_LOCAL_HOST = False
USE_DEV_ENVIRONMENT = False
USE_STAGING_ENVIRONMENT = False

GET_TOKEN_SUFFIX = "/api/user/token"
REGISTER_USER_SUFFIX = "/api/user/register"
GET_DATASET_CREDENTIALS_SUFFIX = "/api/org/{}/ds/{}/creds"
GET_PRESIGNED_URL_SUFFIX = "/api/org/{}/ds/{}/chunks/url/presigned"
CREATE_DATASET_SUFFIX = "/api/dataset/create"
SEND_EVENT_SUFFIX = "/api/event"
DATASET_SUFFIX = "/api/dataset"
UPDATE_SUFFIX = "/api/org/{}/dataset/{}"
GET_MANAGED_CREDS_SUFFIX = "/api/org/{}/storage/name"
ACCEPT_AGREEMENTS_SUFFIX = "/api/organization/{}/dataset/{}/agree"
REJECT_AGREEMENTS_SUFFIX = "/api/organization/{}/dataset/{}/disagree"
LIST_DATASETS = "/api/datasets/{}"
GET_USER_PROFILE = "/api/user/profile"

DEFAULT_REQUEST_TIMEOUT = 170

HUB_AUTH_TOKEN = "HUB_AUTH_TOKEN"
